#=========================================================================
# TranslationImport_stdlib_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 15, 2019
"""Test stdlib RTL components with SystemVerilog translation and import.

We reuse all the test cases from stdlib test files. To achieve this I
overwrite the original reference to the test function used in stdlib test
files and add my own test function (the `run_test`s). However we need
to make sure the orignal reference is not lost and is restored after
finishing each test (no matter it fails or passes).
"""

from pymtl3.passes.backends.sverilog.test.TranslationImport_stdlib_test import (
    test_2entry_normal_Bits,
    test_3entry_normal_Bits,
    test_arbiter_rr_arb_4,
    test_arbiter_rr_arb_en_4,
    test_bypass_Bits,
    test_crossbar3,
    test_encoder_5_directed,
    test_normal_Bits,
    test_pipe_Bits,
)
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test import TestVectorSimulator

from ..TranslationImportPass import TranslationImportPass


def local_do_test( _m ):
  try:
    _m.elaborate()
    # Mark component `_m` as to be translated and imported
    _m.yosys_translate_import = True
    m = TranslationImportPass()( _m )
    sim = TestVectorSimulator( m, _m._test_vectors, _m._tv_in, _m._tv_out )
    sim.run_test()
  finally:
    try:
      # Explicitly finalize imported component to avoid shared lib name aliasing
      m.finalize()
    except UnboundLocalError:
      # This test fails due to translation errors
      pass



"""
========================================================================
bits_import.py
========================================================================
Import RPython Bits from PyPy mamba module if the environment variable
that forces the use of Python Bits is set, and there is actually an
importable Bits in mamba module. Otherwise import the Pure-Python
implementation in Bits.py. Then generate a bunch of fixed-width BitsN
types for PyMTL use.

Author : Shunning Jiang
Date   : Aug 23, 2018
"""
import os

# This __new__ approach has better performance
# bits_template = """
# class Bits{nbits}:
  # nbits = {nbits}
  # def __new__( cls, value = 0 ):
    # return Bits( {nbits}, value )

if os.getenv("PYMTL_BITS") == "1":
  from .PythonBits import Bits
  # print "[env: PYMTL_BITS=1] Use Python Bits"
  bits_template = """
class Bits{0}(Bits):
  nbits = {0}
  def __init__( s, value=0 ):
    return super().__init__( {0}, value )
_bits_types[{nbits}] = b{nbits} = Bits{nbits}
"""
else:
  try:
    from mamba import Bits
    # print "[default w/  Mamba] Use Mamba Bits"
    bits_template = """
class Bits{0}(Bits):
  nbits = {0}
  def __new__( cls, value=0 ):
    return Bits.__new__( cls, {0}, value )
_bits_types[{0}] = b{0} = Bits{0}
"""
  except ImportError:
    from .PythonBits import Bits
    # print "[default w/o Mamba] Use Python Bits"
    bits_template = """
class Bits{0}(Bits):
  nbits = {0}
  def __init__( s, value=0 ):
    return super().__init__( {0}, value )
_bits_types[{0}] = b{0} = Bits{0}
"""

_bitwidths  = list(range(1, 256)) + [ 384, 512 ]
_bits_types = dict()

exec(compile( "".join([ bits_template.format(nbits) for nbits in _bitwidths ]),
              filename="bits_import.py", mode="exec") )

def mk_bits( nbits ):
  # assert nbits < 512, "We don't allow bitwidth to exceed 512."
  if nbits not in _bits_types:
    exec(compile( bits_template.format(nbits), filename=f"Bits{nbits}", mode="exec" ))
  return _bits_types[nbits]



#=========================================================================
# BaseRTLIRTranslator.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 11, 2019
"""Provide base class and metadata namespace for RTLIR translators."""


class BaseRTLIRTranslator:
  """Base class of RTLIR translators."""

  def __init__( s, top ):
    s.top = top

  def clear( s, tr_top ):
    s.tr_top = tr_top
    s.component = {}
    s.hierarchy = TranslatorMetadata()
    s.gen_base_rtlir_trans_metadata( s.tr_top )

  def gen_base_rtlir_trans_metadata( s, m ):
    s.component[m] = TranslatorMetadata()
    for child in m.get_child_components():
      s.gen_base_rtlir_trans_metadata( child )

#-------------------------------------------------------------------------
# TranslatorMetadata
#-------------------------------------------------------------------------

class TranslatorMetadata:
  """Metadata namespace used by RTLIR translators."""
  def __init__( s ):
    pass



#=========================================================================
# ImportedObject_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 2, 2019
"""Test if the imported object works correctly."""

from pymtl3.passes.backends.sverilog.import_.test.ImportedObject_test import (
    test_adder,
    test_reg,
)
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.test import TestVectorSimulator

from ..ImportPass import ImportPass


def local_do_test( _m ):
  _m.elaborate()
  _m.config_yosys_import = _m.config_sverilog_import
  ipass = ImportPass()
  _m.config_yosys_import.fill_missing( _m )
  m = ipass.get_imported_object( _m )
  sim = TestVectorSimulator( m, _m._test_vectors, _m._tv_in, _m._tv_out )
  sim.run_test()



"""
========================================================================
NamedObject_test.py
========================================================================

Author : Shunning Jiang
Date   : Dec 23, 2017
"""
from pymtl3.dsl.NamedObject import NamedObject


class Chicken(NamedObject):

  def construct( s ):
    s.protein = NamedObject()

class Dog(NamedObject):

  def construct( s ):
    s.chicken = Chicken()

class Tiger(NamedObject):

  def construct( s ):
    s.rooster = Chicken()

class Human(NamedObject):

  def construct( s, nlunch=1, ndinner=1 ):

    if nlunch == 1: s.lunch = Tiger()
    else:           s.lunch = [ Tiger() for _ in range(nlunch) ]

    if ndinner == 1: s.dinner = Dog()
    else:            s.dinner = [ Dog() for _ in range(ndinner) ]

def test_NamedObject_normal():

  x = Human( nlunch=1, ndinner=1 )
  x.elaborate()

  assert repr(x) == "s"

  assert repr(x.lunch) == "s.lunch"
  assert repr(x.dinner) == "s.dinner"

  assert repr(x.lunch.rooster) == "s.lunch.rooster"
  assert repr(x.dinner.chicken)== "s.dinner.chicken"

  assert repr(x.lunch.rooster.protein) == "s.lunch.rooster.protein"
  assert repr(x.dinner.chicken.protein)== "s.dinner.chicken.protein"
  print(x.dinner.chicken.protein)

def test_NamedObject_list1():

  x = Human( nlunch=1, ndinner=5 )
  x.elaborate()

  assert repr(x) == "s"

  assert repr(x.lunch) == "s.lunch"
  assert repr(x.dinner[2]) == "s.dinner[2]"

  assert repr(x.lunch.rooster) == "s.lunch.rooster"
  assert repr(x.dinner[2].chicken)== "s.dinner[2].chicken"

  assert repr(x.lunch.rooster.protein) == "s.lunch.rooster.protein"
  assert repr(x.dinner[1].chicken.protein)== "s.dinner[1].chicken.protein"
  print(x.dinner[1].chicken.protein)

def test_NamedObject_list2():

  x = Human( nlunch=4, ndinner=1 )
  x.elaborate()

  assert repr(x) == "s"

  assert repr(x.lunch[3]) == "s.lunch[3]"
  assert repr(x.dinner) == "s.dinner"

  assert repr(x.lunch[3].rooster) == "s.lunch[3].rooster"
  assert repr(x.dinner.chicken) == "s.dinner.chicken"

  assert repr(x.lunch[3].rooster.protein) == "s.lunch[3].rooster.protein"
  assert repr(x.dinner.chicken.protein) == "s.dinner.chicken.protein"
  print(repr(x.lunch[3].rooster.protein))

def test_use_init_error():

  class Crocodile(NamedObject):

    def __init__( s ):
      s.food = Dog()
      print(s.food.chicken)

  # x = Chicken()
  # x.elaborate()
  # y = Crocodile()
  # y.elaborate()
  # z = Chicken()
  # z.elaborate()

def test_set_param():

  class HoneyBadger( NamedObject ):
    def construct( s, lunch="dirt", dinner="dirt" ):
      s.lunch  = lunch
      s.dinner = dinner

  class Dromaius( NamedObject ):
    def construct( s, lunch="dirt", dinner="dirt" ):
      s.lunch  = lunch
      s.dinner = dinner

  class Panda( NamedObject ):
    def construct( s, lunch="dirt", dinner="dirt" ):
      s.lunch  = lunch
      s.dinner = dinner

  class Zoo( NamedObject ):
    def construct( s, AnimalTypes=[] ):
      s.animals = [ A() for A in AnimalTypes ]

  A = Dromaius()
  A.set_param( "top.construct", lunch="grass" )
  A.elaborate()

  print( A.lunch )
  assert A.lunch  == "grass"
  assert A.dinner == "dirt"

  Z = Zoo()
  Z.set_param( "top.construct", AnimalTypes=[ HoneyBadger, Dromaius, Panda ] )
  Z.set_param( "top.animals[0].construct", dinner="bamboo" )
  Z.set_param( "top.animals*.construct",
      lunch ="grass",
      dinner="poisoned onion",
  )
  Z.set_param( "top.animals[2].construct", dinner="bamboo" )
  print( "="*30, "Z", "="*30 )
  print( Z._dsl.param_tree.leaf )
  print( Z._dsl.param_tree.children )
  print( "="*30, "Z.animals[0]", "="*30 )
  print( Z._dsl.param_tree.children["animals[0]"].leaf )
  print( Z._dsl.param_tree.children["animals[0]"].children )
  print( "="*30, "Z.animals*", "="*30 )
  print( Z._dsl.param_tree.children["animals*"].leaf )
  print( Z._dsl.param_tree.children["animals*"].children )
  print( "="*30, "Z.animals[2]", "="*30 )
  print( Z._dsl.param_tree.children["animals[2]"].leaf )
  print( Z._dsl.param_tree.children["animals[2]"].children )
  Z.elaborate()
  print( "="*30, "animals[0]", "="*30 )
  print( Z.animals[0]._dsl.param_tree.leaf )
  print( Z.animals[1]._dsl.param_tree.children )
  print( "="*30, "animals[1]", "="*30 )
  print( Z.animals[0]._dsl.param_tree.leaf )
  print( Z.animals[1]._dsl.param_tree.children )
  print( "="*30, "animals[2]", "="*30 )
  print( Z.animals[0]._dsl.param_tree.leaf )
  print( Z.animals[1]._dsl.param_tree.children )

  assert Z.animals[0].lunch  == "grass"
  assert Z.animals[1].lunch  == "grass"
  assert Z.animals[2].lunch  == "grass"
  assert Z.animals[0].dinner == "poisoned onion"
  assert Z.animals[1].dinner == "poisoned onion"
  assert Z.animals[2].dinner == "bamboo"



#=========================================================================
# YosysBehavioralTranslatorL2.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Provide the yosys-compatible SystemVerilog L2 behavioral translator."""

from pymtl3.passes.backends.sverilog.translation.behavioral.SVBehavioralTranslatorL2 import (
    BehavioralRTLIRToSVVisitorL2,
    SVBehavioralTranslatorL2,
)
from pymtl3.passes.backends.sverilog.util.utility import make_indent

from .YosysBehavioralTranslatorL1 import (
    YosysBehavioralRTLIRToSVVisitorL1,
    YosysBehavioralTranslatorL1,
)


class YosysBehavioralTranslatorL2(
    YosysBehavioralTranslatorL1, SVBehavioralTranslatorL2 ):

  def _get_rtlir2sv_visitor( s ):
    return YosysBehavioralRTLIRToSVVisitorL2

  def rtlir_tr_behavioral_tmpvars( s, tmpvars ):
    _tmpvars = []
    for tmpvar in tmpvars:
      _tmpvars += tmpvar
    make_indent( _tmpvars, 1 )
    return '\n'.join( _tmpvars )

class YosysBehavioralRTLIRToSVVisitorL2(
    YosysBehavioralRTLIRToSVVisitorL1, BehavioralRTLIRToSVVisitorL2 ):

  #-----------------------------------------------------------------------
  # visit_If
  #-----------------------------------------------------------------------

  def visit_If( s, node ):
    node.cond._top_expr = 1
    return super().visit_If( node )

  #-----------------------------------------------------------------------
  # visit_For
  #-----------------------------------------------------------------------

  def visit_For( s, node ):
    node.start._top_expr = 1
    node.end._top_expr = 1
    node.step._top_expr = 1

    # Yosys-comptabile Verilog for loop
    src      = []
    body     = []
    loop_var = s.visit( node.var )
    start    = s.visit( node.start )
    end      = s.visit( node.end )

    loop_var = "__loopvar__" + s.blk.__name__ + "_" + loop_var
    if loop_var not in s.loopvars:
      s.loopvars.add( loop_var )

    begin    = ' begin' if len( node.body ) > 1 else ''

    cmp_op   = '<' if node.step.value > 0 else '<'
    inc_op   = '+' if node.step.value > 0 else '-'

    step_abs = s.visit( node.step )
    step_abs = step_abs if node.step.value > 0 else step_abs[ 1 : ]

    for stmt in node.body:
      body.extend( s.visit( stmt ) )
    make_indent( body, 1 )

    for_begin = \
      'for ( {v} = {s}; {v} {comp} {t}; {v} = {v} {inc} {stp} ){begin}'.format(
      v = loop_var, s = start, t = end, stp = step_abs,
      comp = cmp_op, inc = inc_op, begin = begin
    )

    # Assemble for statement
    src.extend( [ for_begin ] )
    src.extend( body )

    if len( node.body ) > 1:
      src.extend( [ 'end' ] )

    return src

  #-----------------------------------------------------------------------
  # visit_IfExp
  #-----------------------------------------------------------------------

  def visit_IfExp( s, node ):
    node.cond._top_expr = 1
    node.body._top_expr = 1
    node.orelse._top_expr = 1
    return super().visit_IfExp( node )

  #-----------------------------------------------------------------------
  # visit_UnaryOp
  #-----------------------------------------------------------------------

  def visit_UnaryOp( s, node ):
    node.operand._top_expr = 1
    return super().visit_UnaryOp( node )

  #-----------------------------------------------------------------------
  # visit_BoolOp
  #-----------------------------------------------------------------------

  def visit_BoolOp( s, node ):
    for child in node.values:
      child._top_expr = 1
    return super().visit_BoolOp( node )

  #-----------------------------------------------------------------------
  # visit_BinOp
  #-----------------------------------------------------------------------

  def visit_BinOp( s, node ):
    node.left._top_expr = 1
    node.right._top_expr = 1
    return super().visit_BinOp( node )

  #-----------------------------------------------------------------------
  # visit_Compare
  #-----------------------------------------------------------------------

  def visit_Compare( s, node ):
    node.left._top_expr = 1
    node.right._top_expr = 1
    return super().visit_Compare( node )

  #-----------------------------------------------------------------------
  # visit_LoopVar
  #-----------------------------------------------------------------------

  def visit_LoopVar( s, node ):
    s.check_res( node, node.name )
    return "__loopvar__" + s.blk.__name__ + "_" + node.name



"""
======================================================================
Crossbar.py
======================================================================
"""

from pymtl3 import *


class Crossbar( Component ):

  def construct( s, nports, dtype ):

    sel_nbits = clog2( nports )

    s.in_ = [ InPort  ( dtype )     for _ in range( nports ) ]
    s.out = [ OutPort ( dtype )     for _ in range( nports ) ]
    s.sel = [ InPort  ( mk_bits( sel_nbits ) ) for _ in range( nports ) ]

    @s.update
    def comb_logic():

      for i in range( nports ):
        s.out[i] = s.in_[ s.sel[ i ] ]

  def line_trace( s ):
    in_str  = ' '.join( [ str(x) for x in s.in_ ] )
    sel_str = ' '.join( [ str(x) for x in s.sel ] )
    out_str = ' '.join( [ str(x) for x in s.out ] )
    return '{} ( {} ) {}'.format( in_str, sel_str, out_str )



"""
========================================================================
NamedObject.py
========================================================================
This is the very base class of PyMTL objects. It enables all pymtl
components/signals to share the functionality of recursively collecting
objects and tagging objects with the full name.

We bookkeep name hierarchy for error message and other purposes.
For example, s.x[0][3].y[2].z[1:3] is stored as
( ["top","x","y","z"], [ [], [0,3], [2], [slice(1,3,0)] ] )
Note that len(name) == len(idx)-1 only when the variable ends with slice

We keep all metadata in inst._dsl.*. This is to create a namespace
to centralize all DSL-related metadata. Passes will create other
namespaces to put their created metadata.

Author : Shunning Jiang, Yanghui Ou
Date   : Nov 3, 2018
"""
import re
from collections import deque

from .errors import NotElaboratedError


class DSLMetadata:
  pass

# Special data structure for constructing the parameter tree.
class ParamTreeNode:
  def __init__( self ):
    self.compiled_re = None
    self.children = None
    self.leaf = None

  # TODO do we still need to lazily create leaf?
  def merge( self, other ):
    # Merge leaf
    if other.leaf is not None:
      # Lazily create leaf
      if self.leaf is None:
        self.leaf = {}
      for func_name, subdict in other.leaf.items():
        if func_name not in self.leaf:
          self.leaf[ func_name ] = {}
        self.leaf[ func_name ].update( subdict )

    # Merge children
    if other.children is not None:
      # Lazily create children
      if self.children is None:
        self.children = {}
      for comp_name, node in other.children.items():
        if comp_name in self.children:
          self.children[ comp_name ].merge( node )
        else:
          self.children[ comp_name ] = node

  def add_params( self, strs, func_name, **kwargs ):

    if self.leaf is None:
      self.leaf = {}
      self.children = {}

    # Traverse to the node
    cur_node = self
    idx = 1
    for comp_name in strs:
      # Lazily create children
      if cur_node.children is None:
        cur_node.children = {}
      if comp_name not in cur_node.children:
        new_node = ParamTreeNode()
        if '*' in comp_name:
          new_node.compiled_re = re.compile( comp_name )
          # Recursively update exisiting nodes that matches the regex
          for name, node in cur_node.children.items():
            if node.compiled_re is None:
              if new_node.compiled_re.match( name ):
                node.add_params( strs[idx:], func_name, **kwargs )
        cur_node.children[ comp_name ] = new_node
        cur_node = new_node
      else:
        new_node = cur_node.children.pop( comp_name )
        cur_node.children[ comp_name ] = new_node
        cur_node = new_node
      idx += 1

    # Add parameters to leaf
    if cur_node.leaf is None:
      cur_node.leaf = {}
    if func_name not in cur_node.leaf:
      cur_node.leaf[ func_name ] = {}
    cur_node.leaf[ func_name].update( kwargs )

  def __repr__( self ):
    return f"\nleaf:{self.leaf}\nchildren:{self.children}"

class NamedObject:

  def __new__( cls, *args, **kwargs ):

    inst = super().__new__( cls )
    inst._dsl = DSLMetadata() # TODO an actual object?

    # Save parameters for elaborate

    inst._dsl.args        = args
    inst._dsl.kwargs      = kwargs
    inst._dsl.constructed = False

    # A tree of parameters.
    inst._dsl.param_tree = None

    return inst

  #-----------------------------------------------------------------------
  # Private methods
  #-----------------------------------------------------------------------

  def _construct( s ):

    if not s._dsl.constructed:

      # Merge the actual keyword args and those args set by set_parameter
      if s._dsl.param_tree is None:
        kwargs = s._dsl.kwargs
      elif s._dsl.param_tree.leaf is None:
        kwargs = s._dsl.kwargs
      else:
        kwargs = s._dsl.kwargs
        if "construct" in s._dsl.param_tree.leaf:
          more_args = s._dsl.param_tree.leaf[ "construct" ]
          kwargs.update( more_args )

      s.construct( *s._dsl.args, **kwargs )

      s._dsl.constructed = True

  def __setattr_for_elaborate__( s, name, obj ):

    # I use non-recursive BFS to reduce error message depth

    if name[0] != '_': # filter private variables
      sd = s._dsl

      # Shunning: here I optimize for common cases where the object is a NamedObject.
      # I used to push the object directly to a stack to reuse the code across
      # both NamedObject and list cases. Now I basically avoid the stack overheads
      # for common cases.
      if isinstance( obj, NamedObject ):
        ud = obj._dsl

        ud.parent_obj = s
        ud.level      = sd.level + 1

        ud._my_name  = ud.my_name = name
        ud.full_name = f"{sd.full_name}.{name}"

        ud._my_indices = None

        # Iterate through the param_tree and update u
        if sd.param_tree is not None:
          if sd.param_tree.children is not None:
            for comp_name, node in sd.param_tree.children.items():
              if comp_name == name:
                # Lazily create the param tree
                if ud.param_tree is None:
                  ud.param_tree = ParamTreeNode()
                ud.param_tree.merge( node )

              elif node.compiled_re is not None:
                if node.compiled_re.match( name ):
                  # Lazily create the param tree
                  if ud.param_tree is None:
                    ud.param_tree = ParamTreeNode()
                  ud.param_tree.merge( node )

        # Point u's top to my top
        top = sd.elaborate_top
        ud.elaborate_top = top

        top._dsl.elaborate_stack.append( obj )
        obj._construct()
        top._dsl.elaborate_stack.pop()

      # ONLY LIST IS SUPPORTED, SORRY.
      # I don't want to support any iterable object because later "Wire"
      # can be infinitely iterated and cause infinite loop. Special
      # casing Wire will be a mess around everywhere.

      elif isinstance( obj, list ) and obj and isinstance( obj[0], (NamedObject, list) ):
        sd = s._dsl

        Q = deque( (u, (i,)) for i, u in enumerate(obj) )

        while Q:
          u, indices = Q.popleft()

          if isinstance( u, NamedObject ):
            ud = u._dsl

            ud.parent_obj = s
            ud.level      = sd.level + 1

            ud._my_name  = name
            ud.my_name   = u_name = name + "".join( [ f"[{x}]" for x in indices ] )
            ud.full_name = f"{sd.full_name}.{u_name}"

            ud._my_indices = indices

            # Iterate through the param_tree and update u
            if sd.param_tree is not None:
              if sd.param_tree.children is not None:
                for comp_name, node in sd.param_tree.children.items():
                  if comp_name == u_name:
                    # Lazily create the param tree
                    if ud.param_tree is None:
                      ud.param_tree = ParamTreeNode()
                    ud.param_tree.merge( node )

                  elif node.compiled_re is not None:
                    if node.compiled_re.match( u_name ):
                      # Lazily create the param tree
                      if ud.param_tree is None:
                        ud.param_tree = ParamTreeNode()
                      ud.param_tree.merge( node )

            # Point u's top to my top
            top = sd.elaborate_top
            ud.elaborate_top = top

            top._dsl.elaborate_stack.append( u )
            u._construct()
            top._dsl.elaborate_stack.pop()

          elif isinstance( u, list ):
            Q.extend( (v, indices+(i,)) for i, v in enumerate(u) )

    super().__setattr__( name, obj )

  def _collect_all_single( s, filt=lambda x: isinstance( x, NamedObject ) ):
    ret = set()
    stack = [s]
    while stack:
      u = stack.pop()

      if   isinstance( u, NamedObject ):
        if filt( u ): # Check if m satisfies the filter
          ret.add( u )

        for name, obj in u.__dict__.items():

          # If the id is string, it is a normal children field. Otherwise it
          # should be an tuple that represents a slice

          if   isinstance( name, str ):
            if name[0] != '_': # filter private variables
              stack.append( obj )

          elif isinstance( name, tuple ): # name = [1:3]
            stack.append( obj )

      # ONLY LIST IS SUPPORTED
      elif isinstance( u, list ):
        stack.extend( u )
    return ret

  # It is possible to take multiple filters
  def _collect_all( s, filt=[ lambda x: isinstance( x, NamedObject ) ] ):
    ret = [ set() for _ in filt ]
    stack = [s]
    while stack:
      u = stack.pop()
      if   isinstance( u, NamedObject ):

        for i in range( len(filt) ):
          if filt[i]( u ): # Check if m satisfies the filter
            ret[i].add( u )

        for name, obj in u.__dict__.items():

          # If the id is string, it is a normal children field. Otherwise it
          # should be an tuple that represents a slice

          if   isinstance( name, str ):
            if name[0] != '_': # filter private variables
              stack.append( obj )

          elif isinstance( name, tuple ): # name = [1:3]
            stack.append( obj )

      # ONLY LIST IS SUPPORTED
      elif isinstance( u, list ):
        stack.extend( u )
    return ret

  # Developers should use repr(x) everywhere to get the name

  def __repr__( s ):
    try:
      return s._dsl.full_name
    except AttributeError:
      return super().__repr__()

  #-----------------------------------------------------------------------
  # Construction time APIs
  #-----------------------------------------------------------------------

  def construct( s, *args, **kwargs ):
    pass

  def set_param( s, string, **kwargs ):
    # Assert no positional argumets
    # assert len( s._dsl.args ) == 0, \
    #   "Cannot use set_param because {} has positional arguments!".format(s._dsl.my_name)
    assert not s._dsl.constructed

    strs = string.split( "." )

    assert strs[0] == "top", "The component should start at top"
    assert '*' not in strs[-1], "We don't support * with function name!"

    assert len( strs ) >= 2
    func_name = strs[-1]
    strs = strs[1:-1]
    if s._dsl.param_tree is None:
      s._dsl.param_tree = ParamTreeNode()
    s._dsl.param_tree.add_params( strs, func_name, **kwargs )

  # There are two reason I refactored this function into two separate
  # functions. First of all in later levels of components, named objects
  # can be spawned after the previous monolithic elaborate and hence this
  # collect part won't capture them. Second, later levels can override
  # this function and simply call construct at the beginning and call
  # collect at the middle/end.

  #-----------------------------------------------------------------------
  # elaborate
  #-----------------------------------------------------------------------

  def _elaborate_construct( s ):

    if s._dsl.constructed:
      # Yanghui : Mute the warning for the isca tutorial.
      # warnings.warn( "Don't elaborate the same model twice. "
      #                "Use APIs to mutate the model." )
      return

    # Initialize the top level

    s._dsl.parent_obj    = None
    s._dsl.level         = 0
    s._dsl.my_name       = "s"
    s._dsl.full_name     = "s"
    s._dsl.elaborate_top = s

    s._dsl.elaborate_stack = [ s ]

    # Secret source for letting the child know the field name of itself
    # -- override setattr for elaboration, and remove it afterwards

    NamedObject.__setattr__ = NamedObject.__setattr_for_elaborate__

    try:
      s._construct()
    except Exception:
      # re-raise here after deleting __setattr__
      del NamedObject.__setattr__ # not harming the rest of execution
      raise

    del NamedObject.__setattr__

    del s._dsl.elaborate_stack

  def _elaborate_collect_all_named_objects( s ):
    s._dsl.all_named_objects = s._collect_all_single()

  def elaborate( s ):
    s._elaborate_construct()
    s._elaborate_collect_all_named_objects()

  #-----------------------------------------------------------------------
  # Post-elaborate public APIs (can only be called after elaboration)
  #-----------------------------------------------------------------------

  def is_component( s ):
    raise NotImplementedError

  def is_signal( s ):
    raise NotImplementedError

  def is_interface( s ):
    raise NotImplementedError

  # These two APIs are reused across Connectable and Component

  def get_field_name( s ):
    try:
      return s._dsl.my_name
    except AttributeError:
      raise NotElaboratedError()

  def get_parent_object( s ):
    try:
      return s._dsl.parent_obj
    except AttributeError:
      raise NotElaboratedError()



"""
========================================================================
ComponentLevel2.py
========================================================================
At level two we introduce implicit variable constraints.
By default we assume combinational semantics:
- upA reads Wire x while upB writes Wire x ==> upB = WR(x) < RD(x) = upA
When upA is marked as update_ff ==> for all upblks upX that
write/read variables in upA, upA < upX:
- upA = RD(x) < WR(x) = upB and upA = WR(x) < RD(x) = upB

Author : Shunning Jiang
Date   : Apr 16, 2018
"""
import ast
import gc
import inspect
import re
from collections import defaultdict

from pymtl3.datatypes import Bits, is_bitstruct_class

from . import AstHelper
from .ComponentLevel1 import ComponentLevel1
from .Connectable import Connectable, Const, InPort, Interface, OutPort, Signal, Wire
from .ConstraintTypes import RD, WR, U, ValueConstraint
from .errors import (
    InvalidConstraintError,
    InvalidFFAssignError,
    InvalidFuncCallError,
    InvalidIndexError,
    InvalidPlaceholderError,
    InvalidUpblkWriteError,
    MultiWriterError,
    NotElaboratedError,
    PyMTLDeprecationError,
    SignalTypeError,
    UpblkFuncSameNameError,
    VarNotDeclaredError,
)
from .NamedObject import NamedObject
from .Placeholder import Placeholder

compiled_re = re.compile('( *(@|def))')

class ComponentLevel2( ComponentLevel1 ):

  #-----------------------------------------------------------------------
  # Private methods
  #-----------------------------------------------------------------------

  def __new__( cls, *args, **kwargs ):
    inst = super().__new__( cls, *args, **kwargs )

    inst._dsl.update_ff = set()

    # constraint[var] = (sign, func)
    inst._dsl.RD_U_constraints = defaultdict(set)
    inst._dsl.WR_U_constraints = defaultdict(set)
    inst._dsl.name_func = {}

    return inst

  def _cache_func_meta( s, func, is_update_ff, given=None ):
    """ Convention: the source of a function/update block across different
    instances should be the same. You can construct different functions
    based on the condition, but please use different names. This not only
    keeps the caching valid, but also make the code more readable.

    According to the convention, we can cache the information of a
    function in the *class object* to avoid redundant parsing. """
    cls = s.__class__
    try:
      name_info = cls._name_info
      name_rd   = cls._name_rd
      name_wr   = cls._name_wr
      name_fc   = cls._name_fc
    except Exception:
      name_info = cls._name_info = {}
      name_rd   = cls._name_rd  = {}
      name_wr   = cls._name_wr  = {}
      name_fc   = cls._name_fc  = {}

    name = func.__name__

    if name not in name_info:
      if given is None:
        _src, _line = inspect.getsourcelines( func )
        _src = "".join( _src )
        _ast = ast.parse( compiled_re.sub( r'\2', _src ) )

        name_info[ name ] = (False, _src, _line, inspect.getsourcefile( func ), _ast )
      else:
        _src, _ast, _line, _file = given

        name_info[ name ] = ( True, _src, _line, _file, _ast )

      name_rd[ name ]  = _rd   = []
      name_wr[ name ]  = _wr   = []
      name_fc[ name ]  = _fc   = []
      AstHelper.extract_reads_writes_calls( s, func, _ast, is_update_ff, _rd, _wr, _fc )

  def _elaborate_read_write_func( s ):

    # We have parsed AST to extract every read/write variable name.
    # I refactor the process of materializing objects in this function
    # Pass in the func as well for error message

    def extract_obj_from_names( func, names, update_ff=False, is_write=False ):

      def expand_array_index( obj, name_depth, node_depth, idx_depth, idx ):
        """ Find s.x[0][*][2], if index is exhausted, jump back to lookup_variable """

        if idx_depth >= len(idx): # exhausted, go to next level of name
          lookup_variable( obj, name_depth+1, node_depth+1 )
          return

        current_idx = idx[ idx_depth ]

        if current_idx == "*": # special case, materialize all objects
          if isinstance( obj, NamedObject ): # Signal[*] is the signal itself
            objs.add( obj )
          else:
            for i, child in enumerate( obj ):
              expand_array_index( child, name_depth, node_depth, idx_depth+1, idx )

        # Here we try to find the value of free variables in the current
        # component scope.
        # tuple: [x] where x is a closure/global variable
        # slice: [x:y] where x and y are either normal integers or
        #        closure/global variable.
        else:
          if isinstance( current_idx, tuple ):
            is_closure, name = current_idx
            current_idx = _closure[ name ] if is_closure else _globals[ name ]
          elif isinstance( current_idx, slice ):
            start = current_idx.start
            if isinstance( start, tuple ):
              is_closure, name = start
              start = _closure[ name ] if is_closure else _globals[ name ]
            stop  = current_idx.stop
            if isinstance( stop, tuple ):
              is_closure, name = stop
              stop = _closure[ name ] if is_closure else _globals[ name ]
            current_idx = slice(start, stop)

          try:
            child = obj[ current_idx ]
          except TypeError: # cannot convert to integer
            raise VarNotDeclaredError( obj, current_idx, func, s, nodelist[node_depth].lineno )
          except IndexError:
            return
          except AssertionError:
            raise InvalidIndexError( obj, current_idx, func, s, nodelist[node_depth].lineno )

          expand_array_index( child, name_depth, node_depth, idx_depth+1, idx )

      def lookup_variable( obj, name_depth, node_depth ):
        """ Look up the object s.a.b.c in s. Jump to expand_array_index if c[] """
        if obj is None:
          return

        if name_depth >= len(obj_name): # exhausted
          if   isinstance( obj, NamedObject ):
            objs.add( obj )
          elif isinstance( obj, list ) and obj: # Exhaust all the elements in the high-d array
            Q = [ *obj ] # PEP 448 -- see https://stackoverflow.com/a/43220129/6470797
            while Q:
              m = Q.pop()
              if isinstance( m, NamedObject ):
                objs.add( m )
              elif isinstance( m, list ):
                Q.extend( m )
          return

        # still have names
        field, idx = obj_name[ name_depth ]
        try:
          child = getattr( obj, field )
        except AttributeError as e:
          print(e)
          raise VarNotDeclaredError( obj, field, func, s, nodelist[node_depth].lineno )

        if not idx: lookup_variable   ( child, name_depth+1, node_depth+1 )
        else:       expand_array_index( child, name_depth,   node_depth+1, 0, idx )

      """ extract_obj_from_names:
      Here we enumerate names and use the above functions to turn names
      into objects """

      _globals = func.__globals__

      _closure = {}
      for i, var in enumerate( func.__code__.co_freevars ):
        try:  _closure[ var ] = func.__closure__[i].cell_contents
        except ValueError: pass

      all_objs = set()

      for obj_name, nodelist in names:
        if obj_name[0][0] == "s":

          objs = set()
          lookup_variable( s, 1, 1 )
          for obj in objs:
            if not isinstance( obj, Signal ) and is_write:
              raise InvalidUpblkWriteError( s, func, nodelist[0].lineno, obj )
            all_objs.add( obj )

          # Check <<= in update_ff
          if update_ff:
            for x in objs:
              x._dsl.needs_double_buffer = True
              if not x.is_top_level_signal():
                raise InvalidFFAssignError( s, func, nodelist[0].lineno,
                  "has an invalid left value. It needs to be a top level signal, not a slice or a subfield.")
              if not issubclass( x._dsl.Type, Bits ) and not is_bitstruct_class( x._dsl.Type ):
                raise InvalidFFAssignError( s, func, nodelist[0].lineno,
                  "has a wrong type on the left value. "
                  "We only allow <<= on Bits/BitStruct type signals, not {x._dsl.Type}")

        # This is a function call without "s." prefix, check func list
        elif obj_name[0][0] in s._dsl.name_func:
          call = s._dsl.name_func[ obj_name[0][0] ]
          all_objs.add( call )

      return all_objs

    """ elaborate_read_write_func """

    # Access cached data in this component

    cls = s.__class__
    try:
      name_rd, name_wr, name_fc = cls._name_rd, cls._name_wr, cls._name_fc
    except AttributeError: # This component doesn't have update block
      pass

    # what object each astnode corresponds to. You can't have two update
    # blocks in one component that have the same ast.
    s._dsl.func_reads  = {}
    s._dsl.func_writes = {}
    s._dsl.func_calls  = {}
    for name, func in s._dsl.name_func.items():
      s._dsl.func_reads [ func ] = extract_obj_from_names( func, name_rd[ name ] )
      s._dsl.func_writes[ func ] = extract_obj_from_names( func, name_wr[ name ] )
      s._dsl.func_calls [ func ] = extract_obj_from_names( func, name_fc[ name ] )

    s._dsl.upblk_reads  = {}
    s._dsl.upblk_writes = {}
    s._dsl.upblk_calls  = {}
    for name, blk in s._dsl.name_upblk.items():
      s._dsl.upblk_reads [ blk ] = extract_obj_from_names( blk, name_rd[ name ] )
      s._dsl.upblk_writes[ blk ] = extract_obj_from_names( blk, name_wr[ name ],
                                    update_ff = blk in s._dsl.update_ff, is_write=True )
      s._dsl.upblk_calls [ blk ] = extract_obj_from_names( blk, name_fc[ name ] )

  # Override
  def _collect_vars( s, m ):
    super()._collect_vars( m )

    if isinstance( m, ComponentLevel2 ):
      s._dsl.all_update_ff |= m._dsl.update_ff

      for k, k_cons in m._dsl.RD_U_constraints.items():
        s._dsl.all_RD_U_constraints[k] |= k_cons

      for k, k_cons in m._dsl.WR_U_constraints.items():
        s._dsl.all_WR_U_constraints[k] |= k_cons

      # I assume different update blocks will always have different ids
      s._dsl.all_upblk_reads.update( m._dsl.upblk_reads )
      s._dsl.all_upblk_writes.update( m._dsl.upblk_writes )

      for blk, calls in m._dsl.upblk_calls.items():
        s._dsl.all_upblk_calls[ blk ] = calls

        for call in calls:

          # Expand function calls. E.g. upA calls fx, fx calls fy and fz
          # This is invalid: fx calls fy but fy also calls fx
          # To detect this, we need to use dfs and see if the current node
          # has an edge to a previously marked ancestor

          def dfs( u, stk ):
            if u not in m._dsl.func_reads:
              return

            # Add all read/write of funcs to the outermost upblk
            s._dsl.all_upblk_reads [ blk ] |= m._dsl.func_reads[u]
            s._dsl.all_upblk_writes[ blk ] |= m._dsl.func_writes[u]

            for v in m._dsl.func_calls[ u ]:
              if v in caller: # v calls someone else there is a cycle
                raise InvalidFuncCallError( \
                  "In class {}\nThe full call hierarchy:\n - {}{}\nThese function calls form a cycle:\n {}\n{}".format(
                    type(m).__name__, # function's hostobj must be m
                    "\n - ".join( [ "{} calls {}".format( caller[x][0].__name__, x.__name__ )
                                    for x in stk ] ),
                    "\n - {} calls {}".format( u.__name__, v.__name__ ),
                    "\n ".join( [ ">>> {} calls {}".format( caller[x][0].__name__, x.__name__)
                                    for x in stk[caller[v][1]+1: ] ] ),
                    " >>> {} calls {}".format( u.__name__, v.__name__ ) ) )

              caller[ v ] = ( u, len(stk) )
              stk.append( v )
              dfs( v, stk )
              del caller[ v ]
              stk.pop()

          # callee's id: (func, the caller's idx in stk)
          caller = { call: ( blk, 0 ) }
          stk    = [ call ] # for error message
          dfs( call, stk )

  def _uncollect_vars( s, m ):
    super()._uncollect_vars( m )

    if isinstance( m, ComponentLevel2 ):
      s._dsl.all_update_ff -= m._dsl.update_ff

      for k in m._dsl.RD_U_constraints:
        s._dsl.all_RD_U_constraints[k] -= m._dsl.RD_U_constraints[k]
      for k in m._dsl.WR_U_constraints:
        s._dsl.all_WR_U_constraints[k] -= m._dsl.RD_U_constraints[k]

      for k in m._dsl.upblks:
        del s._dsl.all_upblk_reads[k]
        del s._dsl.all_upblk_writes[k]
        del s._dsl.all_upblk_calls[k]

  def _check_upblk_writes( s ):

    write_upblks = defaultdict(set)
    for blk, writes in s._dsl.all_upblk_writes.items():
      for wr in writes:
        write_upblks[ wr ].add( blk )

    for obj, wr_blks in write_upblks.items():
      wr_blks = list(wr_blks)

      if len(wr_blks) > 1:
        raise MultiWriterError( \
        "Multiple update blocks write {}.\n - {}".format( repr(obj),
            "\n - ".join([ x.__name__+" at "+repr(s._dsl.all_upblk_hostobj[x]) \
                           for x in wr_blks ]) ) )

      # See VarConstraintPass.py for full information
      # 1) WR A.b.b.b, A.b.b, A.b, A (detect 2-writer conflict)

      x = obj
      while x.is_signal():
        if x is not obj and x in write_upblks:
          wrx_blks = list(write_upblks[x])

          if wrx_blks[0] != wr_blks[0]:
            raise MultiWriterError( \
            "Two-writer conflict in nested struct/slice. \n - {} (in {})\n - {} (in {})".format(
              repr(x), wrx_blks[0].__name__,
              repr(obj), wr_blks[0].__name__ ) )
        x = x.get_parent_object()

      # 4) WR A.b[1:10], A.b[0:5], A.b[6] (detect 2-writer conflict)

      for x in obj.get_sibling_slices():
        # Recognize overlapped slices
        if x.slice_overlap( obj ) and x in write_upblks:
          wrx_blks = list(write_upblks[x])
          raise MultiWriterError( \
            "Two-writer conflict between sibling slices. \n - {} (in {})\n - {} (in {})".format(
              repr(x), wrx_blks[0].__name__,
              repr(obj), wr_blks[0].__name__ ) )

  def _check_port_in_upblk( s ):

    # Check read first
    for blk, reads in s._dsl.all_upblk_reads.items():

      blk_hostobj = s._dsl.all_upblk_hostobj[ blk ]

      for obj in reads:
        host = obj
        while not isinstance( host, ComponentLevel2 ):
          host = host.get_parent_object() # go to the component

        if   isinstance( obj, (InPort, OutPort) ):  pass
        elif isinstance( obj, Wire ):
          if blk_hostobj != host:
            raise SignalTypeError("""[Type 1] Invalid read to Wire:

- Wire "{}" of {} (class {}) is read in update block
       "{}" of {} (class {}).

  Note: Please only read Wire "x.wire" in x's update block.
        (Or did you intend to declare it as an OutPort?)""" \
          .format(  repr(obj), repr(host), type(host).__name__,
                    blk.__name__, repr(blk_hostobj), type(blk_hostobj).__name__ ) )

    # Then check write
    for blk, writes in s._dsl.all_upblk_writes.items():

      blk_hostobj = s._dsl.all_upblk_hostobj[ blk ]

      for obj in writes:
        host = obj
        while not isinstance( host, ComponentLevel2 ):
          host = host.get_parent_object() # go to the component

      # A continuous assignment is implied when a variable is connected to
      # an input port declaration. This makes assignments to a variable
      # declared as an input port illegal. -- IEEE

        if   isinstance( obj, InPort ):
          if host.get_parent_object() != blk_hostobj:
            raise SignalTypeError("""[Type 2] Invalid write to an input port:

- InPort "{}" of {} (class {}) is written in update block
          "{}" of {} (class {}).

  Note: Please only write to children's InPort "x.y.in", not "x.in", in x's update block.""" \
          .format(  repr(obj), repr(host), type(host).__name__,
                    blk.__name__, repr(host), type(host).__name__ ) )

      # A continuous assignment is implied when a variable is connected to
      # the output port of an instance. This makes procedural or
      # continuous assignments to a variable connected to the output port
      # of an instance illegal. -- IEEE

        elif isinstance( obj, OutPort ):
          if blk_hostobj != host:
            raise SignalTypeError("""[Type 3] Invalid write to output port:

- OutPort \"{}\" of {} (class {}) is written in update block
           \"{}\" of {} (class {}).

  Note: Please only write to OutPort "x.out", not "x.y.out", in x's update block.""" \
          .format(  repr(obj), repr(host), type(host).__name__,
                    blk.__name__, repr(blk_hostobj), type(blk_hostobj).__name__, ) )

      # The case of wire is special. We only allow Wire to be written in
      # the same object. One cannot write this from outside

        elif isinstance( obj, Wire ):
          if blk_hostobj != host:
            raise SignalTypeError("""[Type 4] Invalid write to Wire:

- Wire "{}" of {} (class {}) is written in update block
       "{}" of {} (class {}).

  Note: Please only write to Wire "x.wire" in x's update block.
        (Or did you intend to declare it as an InPort?)""" \
          .format(  repr(obj), repr(host), type(host).__name__,
                    blk.__name__, repr(blk_hostobj), type(blk_hostobj).__name__ ) )

  # TODO rename
  def _check_valid_dsl_code( s ):
    s._check_upblk_writes()
    s._check_port_in_upblk()

  #-----------------------------------------------------------------------
  # Construction-time APIs
  #-----------------------------------------------------------------------

  def func( s, func ): # @s.func is for those functions
    if isinstance( s, Placeholder ):
      raise InvalidPlaceholderError( "Cannot define function "
              "in a placeholder component.".format( func.__name__ ) )
    name = func.__name__
    if name in s._dsl.name_func or name in s._dsl.name_upblk:
      raise UpblkFuncSameNameError( name )

    s._dsl.name_func[ name ] = func
    s._cache_func_meta( func, is_update_ff=False )
    return func

  # Override
  def update( s, blk ):
    super().update( blk )

    s._cache_func_meta( blk, is_update_ff=False ) # add caching of src/ast
    return blk

  def update_on_edge( s, blk ):
    raise PyMTLDeprecationError("\ns.update_on_edge decorator has been deprecated! "
                                "\n- Please use @s.update_ff instead.")

  def update_ff( s, blk ):
    super().update( blk )

    s._dsl.update_ff.add( blk )
    s._cache_func_meta( blk, is_update_ff=True ) # add caching of src/ast
    return blk

  # Override
  def add_constraints( s, *args ): # add RD-U/WR-U constraints
    if isinstance( s, Placeholder ):
      raise InvalidPlaceholderError( "Cannot define constraints "
              "in a placeholder component." )

    for (x0, x1, is_equal) in args:

      if   isinstance( x0, U ) and isinstance( x1, U ): # U & U, same
        assert is_equal == False
        assert (x0.func, x1.func) not in s._dsl.U_U_constraints, \
          "Duplicated constraint"
        s._dsl.U_U_constraints.add( (x0.func, x1.func) )

      elif isinstance( x0, ValueConstraint ) and isinstance( x1, ValueConstraint ):
        raise InvalidConstraintError

      elif isinstance( x0, ValueConstraint ) or isinstance( x1, ValueConstraint ):
        assert is_equal == False
        sign = 1 # RD(x) < U is 1, RD(x) > U is -1
        if isinstance( x1, ValueConstraint ):
          sign = -1
          x0, x1 = x1, x0 # Make sure x0 is RD/WR(...) and x1 is U(...)

        if isinstance( x0, RD ):
          assert (sign, x1.func) not in s._dsl.RD_U_constraints[ x0.var ], \
            "Duplicated constraint"
          s._dsl.RD_U_constraints[ x0.var ].add( (sign, x1.func) )
        else:
          assert (sign, x1.func ) not in s._dsl.WR_U_constraints[ x0.var ], \
            "Duplicated constraint"
          s._dsl.WR_U_constraints[ x0.var ].add( (sign, x1.func) )

  #-----------------------------------------------------------------------
  # elaborate
  #-----------------------------------------------------------------------

  # Override
  def _elaborate_declare_vars( s ):
    super()._elaborate_declare_vars()

    s._dsl.all_update_ff = set()

    s._dsl.all_RD_U_constraints = defaultdict(set)
    s._dsl.all_WR_U_constraints = defaultdict(set)

    # We don't collect func's metadata
    # because every func is local to the component
    s._dsl.all_upblk_reads  = {}
    s._dsl.all_upblk_writes = {}
    s._dsl.all_upblk_calls  = {}

    # Like all_components in level1, although this all_signals is a subset
    # of all_named_objects in NamedObject class, I still maintain it here
    # because we want to avoid redundant isinstance check. I'm going to pay the
    # extra cost of removing from both all_named_objects and all_signals
    # when I delete a signal
    s._dsl.all_signals = set()

  # Override
  def _elaborate_collect_all_vars( s ):
    for c in s._dsl.all_named_objects:
      if isinstance( c, Signal ):
        s._dsl.all_signals.add( c )
      elif isinstance( c, ComponentLevel1 ):
        s._dsl.all_components.add( c )
        s._collect_vars( c )

  # Override
  def elaborate( s ):
    # Don't directly use the base class elaborate anymore
    s._elaborate_construct()

    # First elaborate all functions to spawn more named objects
    for c in s._collect_all_single( lambda s: isinstance( s, ComponentLevel2 ) ):
      c._elaborate_read_write_func()

    s._elaborate_collect_all_named_objects()

    s._elaborate_declare_vars()
    s._elaborate_collect_all_vars()

    s._check_valid_dsl_code()

  #-----------------------------------------------------------------------
  # Post-elaborate public APIs (can only be called after elaboration)
  #-----------------------------------------------------------------------
  # We have moved these implementations to Component.py because the
  # outside world should only use Component.py



#=========================================================================
# BehavioralTranslatorL1_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 20, 2019
"""Test the level 1 behavioral translator."""

import pytest

from ...testcases import (
    CaseBits32ArrayClosureConstruct,
    CaseBits32ClosureConstruct,
    CaseTwoUpblksFreevarsComp,
    CaseTwoUpblksSliceComp,
)
from ..BehavioralTranslatorL1 import BehavioralTranslatorL1
from .TestBehavioralTranslator import mk_TestBehavioralTranslator


@pytest.mark.parametrize(
  'case', [
    CaseBits32ClosureConstruct,
    CaseBits32ArrayClosureConstruct,
    CaseTwoUpblksSliceComp,
    CaseTwoUpblksFreevarsComp,
  ]
)
def test_generic_behavioral_L1( case ):
  m = case.DUT()
  m.elaborate()
  tr = mk_TestBehavioralTranslator(BehavioralTranslatorL1)(m)
  tr.clear( m )
  tr.translate_behavioral( m )
  upblk_src = tr.behavioral.upblk_srcs[m]
  decl_freevars = tr.behavioral.decl_freevars[m]
  assert upblk_src == case.REF_UPBLK
  assert decl_freevars == case.REF_FREEVAR



#=========================================================================
# SVStructuralTranslatorL3.py
#=========================================================================
"""Provide SystemVerilog structural translator implementation."""

from pymtl3.passes.backends.generic.structural.StructuralTranslatorL3 import (
    StructuralTranslatorL3,
)
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt

from ...util.utility import make_indent
from .SVStructuralTranslatorL2 import SVStructuralTranslatorL2


class SVStructuralTranslatorL3(
    SVStructuralTranslatorL2, StructuralTranslatorL3 ):

  #-----------------------------------------------------------------------
  # Declarations
  #-----------------------------------------------------------------------

  def rtlir_tr_interface_port_decls( s, port_decls ):
    return sum( port_decls, [] )

  def rtlir_tr_interface_port_decl( s, m, port_id, port_rtype, port_array_type ):
    def _gen_ifc( id_, ifc, n_dim ):
      if not n_dim:
        ret = []
        all_properties = ifc.get_all_properties_packed()
        for name, _rtype in all_properties:
          if isinstance( _rtype, rt.Array ):
            array_type = _rtype
            rtype = _rtype.get_sub_type()
          else:
            array_type = None
            rtype = _rtype
          ret += s.rtlir_tr_interface_port_decl(
            m, id_+"__"+name, rtype, s.rtlir_tr_unpacked_array_type( array_type ) )
        return ret
      else:
        ret = []
        for i in range( n_dim[0] ):
          ret += _gen_ifc( id_+"__"+str(i), ifc, n_dim[1:] )
        return ret
    # Port id is appended after the template of interface id:
    #   > `{id_}__msg` where `msg` is an attribute of the interface.
    if isinstance( port_rtype, rt.Port ):
      port_dtype = s.rtlir_data_type_translation( m, port_rtype.get_dtype() )
      decl_tmplt = port_rtype.get_direction() + ' ' + \
                   port_dtype['decl'] + '__' + port_id + port_array_type['decl']
      return [ decl_tmplt ]
    else:
      n_dim = port_array_type["n_dim"]
      return _gen_ifc( port_id, port_rtype, n_dim )

  def rtlir_tr_interface_decls( s, ifc_decls ):
    all_decls = sum( ifc_decls, [] )
    make_indent( all_decls, 1 )
    return ',\n'.join( all_decls )

  def rtlir_tr_interface_decl( s, ifc_id, ifc_rtype, array_type, port_decls ):
    def gen_interface_array_decl( ifc_id, ifc_rtype, n_dim, c_n_dim, port_decls ):
      # Fill in the interface id to complete the signal name:
      #   > `in_ifc__0__msg` where `__msg` is filled when calling
      # rtlir_tr_interface_port_decl and `id_` = `in_ifc__0`.
      ret = []
      if not n_dim:
        id_ = ifc_id + c_n_dim
        return [pdecl.format( id_ = id_ ) for pdecl in port_decls]
      else:
        return sum( [ gen_interface_array_decl(
          ifc_id, ifc_rtype, n_dim[1:], c_n_dim+"__"+str(idx), port_decls,
        ) for idx in range( n_dim[0] ) ], [] )
    n_dim = array_type['n_dim']
    return gen_interface_array_decl( ifc_id, ifc_rtype, n_dim, '', port_decls )

  #-----------------------------------------------------------------------
  # Signal operations
  #-----------------------------------------------------------------------

  def rtlir_tr_interface_array_index( s, base_signal, index ):
    return f'{base_signal}__{index}'

  def rtlir_tr_interface_attr( s, base_signal, attr ):
    return f'{base_signal}__{attr}'



"""
========================================================================
test_utils
========================================================================
Helper functions for running unit tests

Author : Shunning Jiang
  Date : Jan 23, 2020
"""

import collections
import re

from pymtl3 import *
from pymtl3.datatypes import is_bitstruct_class
from pymtl3.passes import TracingConfigs
from pymtl3.passes.backends.sverilog import (
    ImportConfigs,
    ImportPass,
    TranslationImportPass,
)

#-------------------------------------------------------------------------
# mk_test_case_table
#-------------------------------------------------------------------------

def mk_test_case_table( raw_test_case_table ):

  # First row in test vectors contains port names

  if isinstance(raw_test_case_table[0],str):
    test_param_names = raw_test_case_table[0].split()
  else:
    test_param_names = raw_test_case_table[0]

  TestCase = collections.namedtuple("TestCase",test_param_names)

  ids = []
  test_cases = []
  for row in raw_test_case_table[1:]:
    ids.append( row[0] )
    test_cases.append( TestCase(*row[1:]) )

  return {
    'ids'      : ids,
    'argnames' : ('test_params'),
    'argvalues' : test_cases,
  }

#------------------------------------------------------------------------------
# TestVectorSimulator
#------------------------------------------------------------------------------
# Now we have the same test vector simulator as pymtl2
# https://github.com/cornell-brg/pymtl/blob/master/pclib/test/TestVectorSimulator.py

class TestVectorSimulator:

  def __init__( self, model, test_vectors,
                set_inputs_func, verify_outputs_func, wait_cycles = 0 ):

    self.model               = model
    self.set_inputs_func     = set_inputs_func
    self.verify_outputs_func = verify_outputs_func
    self.test_vectors        = test_vectors
    self.wait_cycles         = wait_cycles

  def run_test( self ):

    self.model.apply( SimulationPass() )

    self.model.sim_reset()

    print()
    for test_vector in self.test_vectors:

      # Set inputs
      self.set_inputs_func( self.model, test_vector )

      self.model.eval_combinational()

      # Print the line trace
      print(self.model.line_trace())

      # Verify outputs
      self.verify_outputs_func( self.model, test_vector )

      self.model.tick()

def run_sim( model, dump_vcd=None, test_verilog=False, line_trace=True, max_cycles=5000 ):

  # Setup the model

  model.elaborate()

  if dump_vcd:
    model.config_tracing = TracingConfigs( tracing='vcd', vcd_file_name=dump_vcd )

  if test_verilog:
    model.config_sverilog_import = ImportConfigs(
      vl_xinit = test_verilog,
    )
    model.sverilog_translate_import = True

  model = TranslationImportPass()( model )

  # Create a simulator

  model.apply( SimulationPass() )

  # Reset model

  model.sim_reset( print_line_trace=line_trace )

  # Run simulation

  while not model.done() and model.simulated_cycles < max_cycles:
    if line_trace:
      model.print_line_trace()
    model.tick()

  # Force a test failure if we timed out

  assert model.simulated_cycles < max_cycles

  # Extra ticks to make VCD easier to read

  model.tick()
  model.tick()
  model.tick()

class RunTestVectorSimError( Exception ):
  pass

def run_test_vector_sim( model, test_vectors, dump_vcd=None, test_verilog=False, line_trace=True ):

  # First row in test vectors contains port names

  if isinstance(test_vectors[0],str):
    port_names = test_vectors[0].split()
  else:
    port_names = test_vectors[0]

  # Remaining rows contain the actual test vectors

  test_vectors = test_vectors[1:]

  # Setup the model

  model.elaborate()

  if dump_vcd:
    model.config_tracing = TracingConfigs( tracing='vcd', vcd_file_name=dump_vcd )

  if test_verilog:
    model.config_sverilog_import = ImportConfigs(
      vl_xinit = test_verilog,
    )
    model.sverilog_translate_import = True

  model = TranslationImportPass()( model )

  # Create a simulator

  model.apply( SimulationPass() )

  # Reset model

  model.sim_reset( print_line_trace=line_trace )

  # Run the simulation

  row_num = 0

  in_ids  = []
  out_ids = []

  groups = [ None ] * len(port_names)
  types  = [ None ] * len(port_names)

  # Preprocess default type
  # Special case for lists of ports
  # NOTE THAT WE ONLY SUPPORT 1D ARRAY and no interface

  for i, port_full_name in enumerate( port_names ):
    if port_full_name[-1] == "*":
      out_ids.append( i )
      port_name = port_full_name[:-1]
    else:
      in_ids.append( i )
      port_name = port_full_name

    if '[' in port_name:

      # Get tokens of the full name

      m = re.match( r'(\w+)\[(\d+)\]', port_name )

      if not m:
        raise Exception(f"Could not parse port name: {port_name}. "
                        f"Currently we don't support interface or high-D array.")

      groups[i] = g = ( True, m.group(1), int(m.group(2)) )

      # Get type of all the ports
      t = type( getattr( model, g[1] )[ int(g[2]) ] )
      types[i] = None if is_bitstruct_class( t ) else t

    else:
      groups[i] = ( False, port_name )
      t = type( getattr( model, port_name ) )
      types[i] = None if is_bitstruct_class( t ) else t

  for row in test_vectors:
    row_num += 1

    # Apply test inputs

    for i in in_ids:

      in_value = row[i]
      t = types[i]
      if t:
        in_value = t( in_value )

      g = groups[i]
      if g[0]:
        getattr( model, g[1] )[g[2]] = in_value
      else:
        setattr( model, g[1], in_value )

    # Evaluate combinational concurrent blocks

    model.eval_combinational()

    # Display line trace output

    if line_trace:
      model.print_line_trace()

    # Check test outputs

    for i in out_ids:
      ref_value = row[i]
      if ref_value == '?':
        continue

      g = groups[i]
      if g[0]:
        out_value = getattr( model, g[1] )[g[2]]
      else:
        out_value = getattr( model, g[1] )

      if out_value != ref_value:

        error_msg = """
run_test_vector_sim received an incorrect value!
- row number     : {row_number}
- port name      : {port_name}
- expected value : {expected_msg}
- actual value   : {actual_msg}
"""
        raise RunTestVectorSimError( error_msg.format(
          row_number   = row_num,
          port_name    = port_name,
          expected_msg = ref_value,
          actual_msg   = out_value
        ))

    # Tick the simulation

    model.tick()

  # Extra ticks to make VCD easier to read

  model.tick()
  model.tick()
  model.tick()



#=========================================================================
# SVImportSignalGen.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 1, 2019
"""Test the SystemVerilog signal generation of imported component."""

from pymtl3.datatypes import Bits1, Bits32, bitstruct, mk_bits
from pymtl3.dsl import Component, InPort, Interface, OutPort
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..ImportPass import ImportPass


def local_do_test( m ):
  m.elaborate()
  rtype = rt.get_component_ifc_rtlir( m )
  ipass = ImportPass()
  symbols, decls, conns = ipass.gen_signal_decl_py( rtype )
  assert symbols == m._ref_symbols
  assert decls == m._ref_decls
  assert conns == m._ref_conns

def test_port_single( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( mk_bits( 322 ) )
  a = A()
  a._ref_symbols = { 'Bits322' : mk_bits(322) }
  a._ref_decls = [
    "s.in_ = InPort( Bits322 )",
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "connect( s.in_, s.mangled__in_[0:322] )",
    "connect( s.reset, s.mangled__reset[0:1] )"
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "connect( s.in_, s.mangled__in_ )",
    "connect( s.reset, s.mangled__reset )"
  ]
  do_test( a )

def test_port_array( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range( 3 ) ]
  a = A()
  a._ref_symbols = {}
  a._ref_decls = [
    "s.in_ = [ InPort( Bits32 ) for _ in range(3) ]",
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "connect( s.in_[0], s.mangled__in_[0][0:32] )",
    "connect( s.in_[1], s.mangled__in_[1][0:32] )",
    "connect( s.in_[2], s.mangled__in_[2][0:32] )",
    "connect( s.reset, s.mangled__reset[0:1] )"
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "connect( s.in_[0], s.mangled__in___05F_0 )",
    "connect( s.in_[1], s.mangled__in___05F_1 )",
    "connect( s.in_[2], s.mangled__in___05F_2 )",
    "connect( s.reset, s.mangled__reset )"
  ]
  do_test( a )

def test_port_2d_array( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ [ InPort( Bits32 ) for _ in range(2) ] for _ in range(3) ]
  a = A()
  a._ref_symbols = {}
  a._ref_decls = [
    "s.in_ = [ [ InPort( Bits32 ) for _ in range(2) ] for _ in range(3) ]",
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "connect( s.in_[0][0], s.mangled__in_[0][0][0:32] )",
    "connect( s.in_[0][1], s.mangled__in_[0][1][0:32] )",
    "connect( s.in_[1][0], s.mangled__in_[1][0][0:32] )",
    "connect( s.in_[1][1], s.mangled__in_[1][1][0:32] )",
    "connect( s.in_[2][0], s.mangled__in_[2][0][0:32] )",
    "connect( s.in_[2][1], s.mangled__in_[2][1][0:32] )",
    "connect( s.reset, s.mangled__reset[0:1] )"
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "connect( s.in_[0][0], s.mangled__in___05F_0___05F0 )",
    "connect( s.in_[0][1], s.mangled__in___05F_0___05F1 )",
    "connect( s.in_[1][0], s.mangled__in___05F_1___05F0 )",
    "connect( s.in_[1][1], s.mangled__in___05F_1___05F1 )",
    "connect( s.in_[2][0], s.mangled__in___05F_2___05F0 )",
    "connect( s.in_[2][1], s.mangled__in___05F_2___05F1 )",
    "connect( s.reset, s.mangled__reset )"
  ]
  do_test( a )

def test_struct_port_single( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( struct )
  a = A()
  a._ref_symbols = { 'struct' : struct }
  a._ref_decls = [
    "s.in_ = InPort( struct )",
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "@s.update",
    "def in_():",
    "  s.mangled__in_[0:32] = s.in_.foo",
    "  s.mangled__in_[32:64] = s.in_.bar",
    "connect( s.reset, s.mangled__reset[0:1] )"
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "@s.update",
    "def in_():",
    "  s.mangled__in___05F_foo = s.in_.foo",
    "  s.mangled__in___05F_bar = s.in_.bar",
    "connect( s.reset, s.mangled__reset )"
  ]
  do_test( a )

def test_struct_port_array( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  a._ref_symbols = { 'struct' : struct }
  a._ref_decls = [
    "s.in_ = [ InPort( struct ) for _ in range(2) ]",
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "@s.update",
    "def in__LBR_0_RBR_():",
    "  s.mangled__in_[0][0:32] = s.in_[0].foo",
    "  s.mangled__in_[0][32:64] = s.in_[0].bar",
    "@s.update",
    "def in__LBR_1_RBR_():",
    "  s.mangled__in_[1][0:32] = s.in_[1].foo",
    "  s.mangled__in_[1][32:64] = s.in_[1].bar",
    "connect( s.reset, s.mangled__reset[0:1] )"
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "@s.update",
    "def in__LBR_0_RBR_():",
    "  s.mangled__in___05F_0___05Ffoo = s.in_[0].foo",
    "  s.mangled__in___05F_0___05Fbar = s.in_[0].bar",
    "@s.update",
    "def in__LBR_1_RBR_():",
    "  s.mangled__in___05F_1___05Ffoo = s.in_[1].foo",
    "  s.mangled__in___05F_1___05Fbar = s.in_[1].bar",
    "connect( s.reset, s.mangled__reset )"
  ]
  do_test( a )

def test_packed_array_port_array( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: [ [ Bits32 for _ in range(2) ] for _ in range(3) ]
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  a._ref_symbols = { 'struct' : struct }
  a._ref_decls = [
    "s.in_ = [ InPort( struct ) for _ in range(2) ]",
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "@s.update",
    "def in__LBR_0_RBR_():",
    "  s.mangled__in_[0][0:32] = s.in_[0].foo[0][0]",
    "  s.mangled__in_[0][32:64] = s.in_[0].foo[0][1]",
    "  s.mangled__in_[0][64:96] = s.in_[0].foo[1][0]",
    "  s.mangled__in_[0][96:128] = s.in_[0].foo[1][1]",
    "  s.mangled__in_[0][128:160] = s.in_[0].foo[2][0]",
    "  s.mangled__in_[0][160:192] = s.in_[0].foo[2][1]",
    "  s.mangled__in_[0][192:224] = s.in_[0].bar",
    "@s.update",
    "def in__LBR_1_RBR_():",
    "  s.mangled__in_[1][0:32] = s.in_[1].foo[0][0]",
    "  s.mangled__in_[1][32:64] = s.in_[1].foo[0][1]",
    "  s.mangled__in_[1][64:96] = s.in_[1].foo[1][0]",
    "  s.mangled__in_[1][96:128] = s.in_[1].foo[1][1]",
    "  s.mangled__in_[1][128:160] = s.in_[1].foo[2][0]",
    "  s.mangled__in_[1][160:192] = s.in_[1].foo[2][1]",
    "  s.mangled__in_[1][192:224] = s.in_[1].bar",
    "connect( s.reset, s.mangled__reset[0:1] )"
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "@s.update",
    "def in__LBR_0_RBR_():",
    "  s.mangled__in___05F_0___05Ffoo___05F0___05F0 = s.in_[0].foo[0][0]",
    "  s.mangled__in___05F_0___05Ffoo___05F0___05F1 = s.in_[0].foo[0][1]",
    "  s.mangled__in___05F_0___05Ffoo___05F1___05F0 = s.in_[0].foo[1][0]",
    "  s.mangled__in___05F_0___05Ffoo___05F1___05F1 = s.in_[0].foo[1][1]",
    "  s.mangled__in___05F_0___05Ffoo___05F2___05F0 = s.in_[0].foo[2][0]",
    "  s.mangled__in___05F_0___05Ffoo___05F2___05F1 = s.in_[0].foo[2][1]",
    "  s.mangled__in___05F_0___05Fbar = s.in_[0].bar",
    "@s.update",
    "def in__LBR_1_RBR_():",
    "  s.mangled__in___05F_1___05Ffoo___05F0___05F0 = s.in_[1].foo[0][0]",
    "  s.mangled__in___05F_1___05Ffoo___05F0___05F1 = s.in_[1].foo[0][1]",
    "  s.mangled__in___05F_1___05Ffoo___05F1___05F0 = s.in_[1].foo[1][0]",
    "  s.mangled__in___05F_1___05Ffoo___05F1___05F1 = s.in_[1].foo[1][1]",
    "  s.mangled__in___05F_1___05Ffoo___05F2___05F0 = s.in_[1].foo[2][0]",
    "  s.mangled__in___05F_1___05Ffoo___05F2___05F1 = s.in_[1].foo[2][1]",
    "  s.mangled__in___05F_1___05Fbar = s.in_[1].bar",
    "connect( s.reset, s.mangled__reset )"
  ]
  do_test( a )

def test_nested_struct( do_test ):
  @bitstruct
  class inner_struct:
    foo: Bits32
  @bitstruct
  class struct:
    bar: Bits32
    inner: inner_struct
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  # Inner struct will not be added to `symbols` because struct
  # refers to it!
  a._ref_symbols = { 'struct' : struct }
  a._ref_decls = [
    "s.in_ = [ InPort( struct ) for _ in range(2) ]",
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "@s.update",
    "def in__LBR_0_RBR_():",
    "  s.mangled__in_[0][0:32] = s.in_[0].inner.foo",
    "  s.mangled__in_[0][32:64] = s.in_[0].bar",
    "@s.update",
    "def in__LBR_1_RBR_():",
    "  s.mangled__in_[1][0:32] = s.in_[1].inner.foo",
    "  s.mangled__in_[1][32:64] = s.in_[1].bar",
    "connect( s.reset, s.mangled__reset[0:1] )"
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "@s.update",
    "def in__LBR_0_RBR_():",
    "  s.mangled__in___05F_0___05Finner___05Ffoo = s.in_[0].inner.foo",
    "  s.mangled__in___05F_0___05Fbar = s.in_[0].bar",
    "@s.update",
    "def in__LBR_1_RBR_():",
    "  s.mangled__in___05F_1___05Finner___05Ffoo = s.in_[1].inner.foo",
    "  s.mangled__in___05F_1___05Fbar = s.in_[1].bar",
    "connect( s.reset, s.mangled__reset )"
  ]
  do_test( a )

def test_interface( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = Ifc()
  a = A()
  a._ref_symbols = { 'Ifc' : Ifc }
  a._ref_decls = [
    "s.ifc = Ifc()"
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "connect( s.reset, s.mangled__reset[0:1] )",
    "connect( s.ifc.msg, s.mangled__ifc___05Fmsg[0:32] )",
    "connect( s.ifc.rdy, s.mangled__ifc___05Frdy[0:1] )",
    "connect( s.ifc.val, s.mangled__ifc___05Fval[0:1] )",
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "connect( s.reset, s.mangled__reset )",
    "connect( s.ifc.msg, s.mangled__ifc___05Fmsg )",
    "connect( s.ifc.rdy, s.mangled__ifc___05Frdy )",
    "connect( s.ifc.val, s.mangled__ifc___05Fval )",
  ]
  do_test( a )

def test_interface_parameter( do_test ):
  class Ifc( Interface ):
    def construct( s, Type, nbits_val, nbits_rdy ):
      s.msg = InPort( Type )
      s.val = InPort( mk_bits(nbits_val) )
      # Added support for BitsN values in case someone wants to do
      # tricky things like this.
      s.rdy = OutPort( mk_bits(nbits_rdy.nbits) )
  class A( Component ):
    def construct( s ):
      s.ifc = Ifc( Bits32, 1, Bits1(1) )
  a = A()
  a._ref_symbols = { 'Ifc' : Ifc }
  a._ref_decls = [
    "s.ifc = Ifc( Bits32, 1, Bits1( 1 ) )"
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "connect( s.reset, s.mangled__reset[0:1] )",
    "connect( s.ifc.msg, s.mangled__ifc___05Fmsg[0:32] )",
    "connect( s.ifc.rdy, s.mangled__ifc___05Frdy[0:1] )",
    "connect( s.ifc.val, s.mangled__ifc___05Fval[0:1] )",
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "connect( s.reset, s.mangled__reset )",
    "connect( s.ifc.msg, s.mangled__ifc___05Fmsg )",
    "connect( s.ifc.rdy, s.mangled__ifc___05Frdy )",
    "connect( s.ifc.val, s.mangled__ifc___05Fval )",
  ]
  do_test( a )

def test_interface_parameter_long_vector( do_test ):
  class Ifc( Interface ):
    def construct( s, Type, nbits_val, nbits_rdy ):
      s.msg = InPort( Type )
      s.val = InPort( mk_bits(nbits_val) )
      # Added support for BitsN values in case someone wants to do
      # tricky things like this.
      s.rdy = OutPort( mk_bits(nbits_rdy.nbits) )
  class A( Component ):
    def construct( s ):
      Bits322 = mk_bits(322)
      s.ifc = Ifc( Bits322, 1, Bits322(1) )
  a = A()
  a._ref_symbols = { 'Ifc' : Ifc, 'Bits322' : mk_bits(322) }
  a._ref_decls = [
    "s.ifc = Ifc( Bits322, 1, Bits322( 1 ) )"
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "connect( s.reset, s.mangled__reset[0:1] )",
    "connect( s.ifc.msg, s.mangled__ifc___05Fmsg[0:322] )",
    "connect( s.ifc.rdy, s.mangled__ifc___05Frdy[0:322] )",
    "connect( s.ifc.val, s.mangled__ifc___05Fval[0:1] )",
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "connect( s.reset, s.mangled__reset )",
    "connect( s.ifc.msg, s.mangled__ifc___05Fmsg )",
    "connect( s.ifc.rdy, s.mangled__ifc___05Frdy )",
    "connect( s.ifc.val, s.mangled__ifc___05Fval )",
  ]
  do_test( a )

def test_interface_array( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  a._ref_symbols = { 'Ifc' : Ifc }
  a._ref_decls = [
    "s.ifc = [ Ifc() for _ in range(2) ]"
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "connect( s.reset, s.mangled__reset[0:1] )",
    "connect( s.ifc[0].msg, s.mangled__ifc___05F0___05Fmsg[0:32] )",
    "connect( s.ifc[0].rdy, s.mangled__ifc___05F0___05Frdy[0:1] )",
    "connect( s.ifc[0].val, s.mangled__ifc___05F0___05Fval[0:1] )",
    "connect( s.ifc[1].msg, s.mangled__ifc___05F1___05Fmsg[0:32] )",
    "connect( s.ifc[1].rdy, s.mangled__ifc___05F1___05Frdy[0:1] )",
    "connect( s.ifc[1].val, s.mangled__ifc___05F1___05Fval[0:1] )"
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "connect( s.reset, s.mangled__reset )",
    "connect( s.ifc[0].msg, s.mangled__ifc___05F0___05Fmsg )",
    "connect( s.ifc[0].rdy, s.mangled__ifc___05F0___05Frdy )",
    "connect( s.ifc[0].val, s.mangled__ifc___05F0___05Fval )",
    "connect( s.ifc[1].msg, s.mangled__ifc___05F1___05Fmsg )",
    "connect( s.ifc[1].rdy, s.mangled__ifc___05F1___05Frdy )",
    "connect( s.ifc[1].val, s.mangled__ifc___05F1___05Fval )"
  ]
  do_test( a )

def test_nested_interface( do_test ):
  class InnerIfc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class Ifc( Interface ):
    def construct( s ):
      s.valrdy_ifc = InnerIfc()
      s.ctrl_bar = InPort( Bits32 )
      s.ctrl_foo = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  # Inner interface will not be added to `symbols` because Ifc refers
  # to it!
  a._ref_symbols = { 'Ifc' : Ifc }
  a._ref_decls = [
    "s.ifc = [ Ifc() for _ in range(2) ]"
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "connect( s.reset, s.mangled__reset[0:1] )",
    "connect( s.ifc[0].ctrl_bar, s.mangled__ifc___05F0___05Fctrl_bar[0:32] )",
    "connect( s.ifc[0].ctrl_foo, s.mangled__ifc___05F0___05Fctrl_foo[0:32] )",
    "connect( s.ifc[0].valrdy_ifc.msg, s.mangled__ifc___05F0___05Fvalrdy_ifc___05Fmsg[0:32] )",
    "connect( s.ifc[0].valrdy_ifc.rdy, s.mangled__ifc___05F0___05Fvalrdy_ifc___05Frdy[0:1] )",
    "connect( s.ifc[0].valrdy_ifc.val, s.mangled__ifc___05F0___05Fvalrdy_ifc___05Fval[0:1] )",
    "connect( s.ifc[1].ctrl_bar, s.mangled__ifc___05F1___05Fctrl_bar[0:32] )",
    "connect( s.ifc[1].ctrl_foo, s.mangled__ifc___05F1___05Fctrl_foo[0:32] )",
    "connect( s.ifc[1].valrdy_ifc.msg, s.mangled__ifc___05F1___05Fvalrdy_ifc___05Fmsg[0:32] )",
    "connect( s.ifc[1].valrdy_ifc.rdy, s.mangled__ifc___05F1___05Fvalrdy_ifc___05Frdy[0:1] )",
    "connect( s.ifc[1].valrdy_ifc.val, s.mangled__ifc___05F1___05Fvalrdy_ifc___05Fval[0:1] )",
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "connect( s.reset, s.mangled__reset )",
    "connect( s.ifc[0].ctrl_bar, s.mangled__ifc___05F0___05Fctrl_bar )",
    "connect( s.ifc[0].ctrl_foo, s.mangled__ifc___05F0___05Fctrl_foo )",
    "connect( s.ifc[0].valrdy_ifc.msg, s.mangled__ifc___05F0___05Fvalrdy_ifc___05Fmsg )",
    "connect( s.ifc[0].valrdy_ifc.rdy, s.mangled__ifc___05F0___05Fvalrdy_ifc___05Frdy )",
    "connect( s.ifc[0].valrdy_ifc.val, s.mangled__ifc___05F0___05Fvalrdy_ifc___05Fval )",
    "connect( s.ifc[1].ctrl_bar, s.mangled__ifc___05F1___05Fctrl_bar )",
    "connect( s.ifc[1].ctrl_foo, s.mangled__ifc___05F1___05Fctrl_foo )",
    "connect( s.ifc[1].valrdy_ifc.msg, s.mangled__ifc___05F1___05Fvalrdy_ifc___05Fmsg )",
    "connect( s.ifc[1].valrdy_ifc.rdy, s.mangled__ifc___05F1___05Fvalrdy_ifc___05Frdy )",
    "connect( s.ifc[1].valrdy_ifc.val, s.mangled__ifc___05F1___05Fvalrdy_ifc___05Fval )",
  ]
  do_test( a )

def test_nested_interface_port_array( do_test ):
  class InnerIfc( Interface ):
    def construct( s ):
      s.msg = [ InPort( Bits32 ) for _ in range(2) ]
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class Ifc( Interface ):
    def construct( s ):
      s.valrdy_ifc = InnerIfc()
      s.ctrl_bar = InPort( Bits32 )
      s.ctrl_foo = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  # Inner interface will not be added to `symbols` because Ifc refers
  # to it!
  a._ref_symbols = { 'Ifc' : Ifc }
  a._ref_decls = [
    "s.ifc = [ Ifc() for _ in range(2) ]"
  ]
  a._ref_conns = [
    "connect( s.clk, s.mangled__clk[0:1] )",
    "connect( s.reset, s.mangled__reset[0:1] )",
    "connect( s.ifc[0].ctrl_bar, s.mangled__ifc___05F0___05Fctrl_bar[0:32] )",
    "connect( s.ifc[0].ctrl_foo, s.mangled__ifc___05F0___05Fctrl_foo[0:32] )",
    "connect( s.ifc[0].valrdy_ifc.msg[0], s.mangled__ifc___05F0___05Fvalrdy_ifc___05Fmsg[0][0:32] )",
    "connect( s.ifc[0].valrdy_ifc.msg[1], s.mangled__ifc___05F0___05Fvalrdy_ifc___05Fmsg[1][0:32] )",
    "connect( s.ifc[0].valrdy_ifc.rdy, s.mangled__ifc___05F0___05Fvalrdy_ifc___05Frdy[0:1] )",
    "connect( s.ifc[0].valrdy_ifc.val, s.mangled__ifc___05F0___05Fvalrdy_ifc___05Fval[0:1] )",
    "connect( s.ifc[1].ctrl_bar, s.mangled__ifc___05F1___05Fctrl_bar[0:32] )",
    "connect( s.ifc[1].ctrl_foo, s.mangled__ifc___05F1___05Fctrl_foo[0:32] )",
    "connect( s.ifc[1].valrdy_ifc.msg[0], s.mangled__ifc___05F1___05Fvalrdy_ifc___05Fmsg[0][0:32] )",
    "connect( s.ifc[1].valrdy_ifc.msg[1], s.mangled__ifc___05F1___05Fvalrdy_ifc___05Fmsg[1][0:32] )",
    "connect( s.ifc[1].valrdy_ifc.rdy, s.mangled__ifc___05F1___05Fvalrdy_ifc___05Frdy[0:1] )",
    "connect( s.ifc[1].valrdy_ifc.val, s.mangled__ifc___05F1___05Fvalrdy_ifc___05Fval[0:1] )",
  ]
  a._ref_conns_yosys = [
    "connect( s.clk, s.mangled__clk )",
    "connect( s.reset, s.mangled__reset )",
    "connect( s.ifc[0].ctrl_bar, s.mangled__ifc___05F0___05Fctrl_bar )",
    "connect( s.ifc[0].ctrl_foo, s.mangled__ifc___05F0___05Fctrl_foo )",
    "connect( s.ifc[0].valrdy_ifc.msg[0], s.mangled__ifc___05F0___05Fvalrdy_ifc___05Fmsg___05F0 )",
    "connect( s.ifc[0].valrdy_ifc.msg[1], s.mangled__ifc___05F0___05Fvalrdy_ifc___05Fmsg___05F1 )",
    "connect( s.ifc[0].valrdy_ifc.rdy, s.mangled__ifc___05F0___05Fvalrdy_ifc___05Frdy )",
    "connect( s.ifc[0].valrdy_ifc.val, s.mangled__ifc___05F0___05Fvalrdy_ifc___05Fval )",
    "connect( s.ifc[1].ctrl_bar, s.mangled__ifc___05F1___05Fctrl_bar )",
    "connect( s.ifc[1].ctrl_foo, s.mangled__ifc___05F1___05Fctrl_foo )",
    "connect( s.ifc[1].valrdy_ifc.msg[0], s.mangled__ifc___05F1___05Fvalrdy_ifc___05Fmsg___05F0 )",
    "connect( s.ifc[1].valrdy_ifc.msg[1], s.mangled__ifc___05F1___05Fvalrdy_ifc___05Fmsg___05F1 )",
    "connect( s.ifc[1].valrdy_ifc.rdy, s.mangled__ifc___05F1___05Fvalrdy_ifc___05Frdy )",
    "connect( s.ifc[1].valrdy_ifc.val, s.mangled__ifc___05F1___05Fvalrdy_ifc___05Fval )",
  ]
  do_test( a )



#=========================================================================
# YosysStructuralTranslatorL1.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Provide the yosys-compatible SystemVerilog structural translator."""

from collections import deque

from pymtl3.passes.backends.sverilog.errors import SVerilogTranslationError
from pymtl3.passes.backends.sverilog.translation.structural.SVStructuralTranslatorL1 import (
    SVStructuralTranslatorL1,
)
from pymtl3.passes.backends.sverilog.util.utility import make_indent
from pymtl3.passes.rtlir import RTLIRDataType as rdt


class YosysStructuralTranslatorL1( SVStructuralTranslatorL1 ):

  def __init__( s, top ):
    super().__init__( top )
    s.deq = deque()

  #-----------------------------------------------------------------------
  # Connection helper method
  #-----------------------------------------------------------------------

  def vector_conn_gen( s, d, pid, wid, idx, dtype ):
    return [ {
      "direction" : d,
      "pid" : pid,
      "wid" : wid,
      "idx" : idx
    } ]

  def dtype_conn_gen( s, d, pid, wid, idx, dtype ):
    assert isinstance( dtype, rdt.Vector )
    return s.vector_conn_gen( d, pid, wid, idx, dtype )

  def _port_conn_gen( s, d, pid, wid, idx, n_dim, dtype ):
    if not n_dim:
      return s.dtype_conn_gen( d, pid, wid, idx, dtype )
    else:
      ret = []
      for i in range( n_dim[0] ):
        _pid = f"{pid}__{i}"
        _idx = f"{idx}[{i}]"
        ret += s._port_conn_gen( d, _pid, wid, _idx, n_dim[1:], dtype )
      return ret

  def port_connection_gen( s, d, id_, n_dim, dtype ):
    return s._port_conn_gen( d, id_, id_, "", n_dim, dtype )

  #-----------------------------------------------------------------------
  # Port wire declaration helper method
  #-----------------------------------------------------------------------

  def _get_array_dim_str( s, n_dim ):
    ret = "".join( f"[0:{size-1}]" for size in n_dim )
    if ret:
      ret = " " + ret
    return ret

  def wire_vector_gen( s, id_, dtype, n_dim ):
    msb = dtype.get_length() - 1
    return [ {
      "msb" : msb,
      "id_" : id_,
      "n_dim" : n_dim
    } ]

  def wire_dtype_gen( s, id_, dtype, n_dim ):
    assert isinstance( dtype, rdt.Vector )
    s.check_decl( id_, "" )
    return s.wire_vector_gen( id_, dtype, n_dim )

  def port_wire_gen( s, id_, n_dim, dtype ):
    return s.wire_dtype_gen( id_, dtype, n_dim )

  #-----------------------------------------------------------------------
  # Port declaration helper methods
  #-----------------------------------------------------------------------

  def vector_gen( s, d, id_, dtype ):
    direction = d + " " if d else ""
    msb = dtype.get_length() - 1
    return [ {
      "direction" : direction,
      "msb" : msb,
      "id_" : id_
    } ]

  def dtype_gen( s, d, id_, dtype ):
    assert isinstance( dtype, rdt.Vector ), f"unrecognized data type {dtype}"
    s.check_decl( id_, "" )
    return s.vector_gen( d, id_, dtype )

  def port_gen( s, d, id_, n_dim, dtype ):
    if not n_dim:
      return s.dtype_gen( d, id_, dtype )
    else:
      ret = []
      for idx in range(n_dim[0]):
        ret += s.port_gen(d, f"{id_}__{idx}", n_dim[1:], dtype)
      return ret

  #---------------------------------------------------------------------
  # Declarations
  #---------------------------------------------------------------------

  def rtlir_tr_port_decls( s, port_decls ):
    ret = { "port_decls" : [], "wire_decls" : [], "connections" : [] }
    for port_decl in port_decls:
      ret["port_decls"] += port_decl["port_decls"]
      ret["wire_decls"] += port_decl["wire_decls"]
      ret["connections"] += port_decl["connections"]
    make_indent( ret["port_decls"], 1 )
    make_indent( ret["wire_decls"], 1 )
    make_indent( ret["connections"], 1 )
    ret["port_decls"] = ",\n".join( ret["port_decls"] )
    ret["wire_decls"] = "\n".join( ret["wire_decls"] )
    ret["connections"] = "\n".join( ret["connections"] )
    return ret

  def rtlir_tr_port_decl( s, port_id, Type, array_type, _dtype ):
    port_template = "{direction: <7}logic {packed_type: <8} {id_}"
    wire_template = "logic {packed_type: <8} {id_}{array_dim_str};"
    in_conn_template = "assign {wid}{idx} = {pid};"
    out_conn_template = "assign {pid} = {wid}{idx};"

    port_dtype = _dtype['raw_dtype']
    port_direction = Type.get_direction()
    port_n_dim = array_type["n_dim"]
    port_decl, wire_decl, connections = [], [], []

    # Assemble port declarations
    decl_list = s.port_gen( port_direction, port_id, port_n_dim, port_dtype )
    for dct in decl_list:
      direction = dct["direction"]
      msb, id_ = dct["msb"], dct["id_"]
      packed_type = f"[{msb}:0]"
      port_decl.append( port_template.format( **locals() ) )

    # Assemble wire declarations
    decl_list = s.port_wire_gen( port_id, port_n_dim, port_dtype )
    for dct in decl_list:
      msb, id_, n_dim = dct["msb"], dct["id_"], dct["n_dim"]
      array_dim_str = s._get_array_dim_str( n_dim )
      packed_type = f"[{msb}:0]"
      if n_dim or "present" in dct:
        wire_decl.append( wire_template.format( **locals() ) )

    # Assemble connections
    conn_list = s.port_connection_gen( port_direction, port_id, port_n_dim, port_dtype )
    for dct in conn_list:
      direction = dct["direction"]
      pid, wid, idx = dct["pid"], dct["wid"], dct["idx"]
      if idx or "present" in dct:
        if direction.startswith( "input" ):
          connections.append( in_conn_template.format( **locals() ) )
        else:
          connections.append( out_conn_template.format( **locals() ) )

    return {
      "port_decls" : port_decl,
      "wire_decls" : wire_decl,
      "connections" : connections
    }

  def rtlir_tr_wire_decls( s, wire_decls ):
    wires = []
    for wire_decl in wire_decls:
      wires += wire_decl
    make_indent( wires, 1 )
    return '\n'.join( wires )

  def rtlir_tr_wire_decl( s, wire_id, Type, array_type, _dtype ):
    wire_template = "logic {packed_type: <8} {id_}{array_dim_str};"

    wire_dtype = _dtype['raw_dtype']
    wire_n_dim = array_type['n_dim']
    wire_decl = []

    # Assemble wire declarations
    decl_list = s.port_wire_gen( wire_id, wire_n_dim, wire_dtype )
    for dct in decl_list:
      msb, id_ = dct["msb"], dct["id_"]
      array_dim_str = s._get_array_dim_str( dct["n_dim"] )
      packed_type = f"[{msb}:0]"
      wire_decl.append( wire_template.format( **locals() ) )

    return wire_decl

  def rtlir_tr_const_decls( s, const_decls ):
    return ""

  def rtlir_tr_const_decl( s, id_, Type, array_type, dtype, value ):
    return ""

  #---------------------------------------------------------------------
  # Connections
  #---------------------------------------------------------------------

  def rtlir_tr_connection( s, wr_signal, rd_signal ):
    # First assemble the WR signal
    sexp = s.deq.popleft()
    all_terms = sexp['attr'] + sexp['index']
    template = sexp['s_attr'] + sexp['s_index']
    wr = template.format( *all_terms )
    # Then assemble the RD signal
    sexp = s.deq.popleft()
    all_terms = sexp['attr'] + sexp['index']
    template = sexp['s_attr'] + sexp['s_index']
    rd = template.format( *all_terms )
    return f"assign {rd} = {wr};"

  #-----------------------------------------------------------------------
  # Signal operations
  #-----------------------------------------------------------------------

  def rtlir_tr_bit_selection( s, base_signal, index ):
    # Bit selection
    s.deq[-1]['s_index'] += "[{}]"
    s.deq[-1]['index'].append( int(index) )
    return f'{base_signal}[{index}]'

  def rtlir_tr_part_selection( s, base_signal, start, stop ):
    # Part selection
    _stop = stop-1
    s.deq[-1]['s_index'] += "[{}:{}]"
    s.deq[-1]['index'].append( int(_stop) )
    s.deq[-1]['index'].append( int(start) )
    return f'{base_signal}[{_stop}:{start}]'

  def rtlir_tr_port_array_index( s, base_signal, index ):
    s.deq[-1]['s_index'] += "[{}]"
    s.deq[-1]['index'].append( int(index) )
    return f'{base_signal}[{index}]'

  def rtlir_tr_wire_array_index( s, base_signal, index ):
    s.deq[-1]['s_index'] += "[{}]"
    s.deq[-1]['index'].append( int(index) )
    return f'{base_signal}[{index}]'

  def rtlir_tr_const_array_index( s, base_signal, index ):
    assert False, f"constant array {base_signal} is not allowed!"

  def rtlir_tr_current_comp_attr( s, base_signal, attr ):
    s.deq[-1]['s_attr'] = attr
    return f'{attr}'

  def rtlir_tr_current_comp( s, comp_id, comp_rtype ):
    s.deq.append( {'attr':[], 'index':[], 's_attr':"", 's_index':""} )
    return ''

  def rtlir_tr_literal_number( s, nbits, value, first_called = True ):
    num_str = s._literal_number( nbits, value )
    if first_called:
      s.deq.append( {'attr':[], 'index':[], 's_attr':num_str, 's_index':""} )
    return num_str



"""
========================================================================
ConstraintType.py
========================================================================
Explicit constraint class types.

Author : Shunning Jiang
Date   : July 7, 2017
"""

class FuncConstraint:
  def __init__( self, func ): self.func = func
  def __lt__( self, other ):  return (self, other, False)
  def __gt__( self, other ):  return (other, self, False)
  def __eq__( self, other ):  return (self, other, True)

class ValueConstraint:
  def __init__( self, var ):  self.var = var
  def __lt__( self, other ):  return (self, other, False)
  def __gt__( self, other ):  return (other, self, False)

class U(FuncConstraint): pass
class M(FuncConstraint): pass

class RD(ValueConstraint): pass
class WR(ValueConstraint): pass



#=========================================================================
# StructuralRTLIRGenL2Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Apr 3, 2019
"""Provide L2 structural RTLIR generation pass."""

from .StructuralRTLIRGenL1Pass import StructuralRTLIRGenL1Pass


class StructuralRTLIRGenL2Pass( StructuralRTLIRGenL1Pass ):
  pass



#=========================================================================
# TranslationImport_closed_loop_directed_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 5, 2019
"""Closed-loop test with SystemVerilog translation and import."""

from pymtl3.datatypes import Bits1, mk_bits
from pymtl3.passes.PassGroups import SimulationPass
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.stdlib.rtl.enrdy_queues_test import test_bypass_queue as _bypass_queue
from pymtl3.stdlib.rtl.enrdy_queues_test import (
    test_bypass_queue_stall as _bypass_queue_stall,
)
from pymtl3.stdlib.rtl.enrdy_queues_test import test_normal_queue as _normal_queue
from pymtl3.stdlib.rtl.enrdy_queues_test import (
    test_normal_queue_stall as _normal_queue_stall,
)
from pymtl3.stdlib.rtl.enrdy_queues_test import test_pipe_queue as _pipe_queue
from pymtl3.stdlib.rtl.enrdy_queues_test import (
    test_pipe_queue_stall as _pipe_queue_stall,
)

from .. import TranslationImportPass
from ..util.test_utility import closed_loop_component_input_test

#-------------------------------------------------------------------------
# Valrdy queue tests
#-------------------------------------------------------------------------

def run_sim( _th ):
  try:
    _th.elaborate()
    _th.q.sverilog_translate_import = True
    th = TranslationImportPass()( _th )
    th.apply( SimulationPass() )

    print()
    cycle = 0
    while not th.done() and cycle < 1000:
      th.tick()
      print(th.line_trace())
      cycle += 1

    assert cycle < 1000

    th.tick()
    th.tick()
    th.tick()
  finally:
    try:
      th.q.finalize()
    except UnboundLocalError:
      # This test fails due to translation errors
      pass

def test_normal_queue():
  test_func = _normal_queue
  _run_test = test_func.__globals__['run_sim']
  test_func.__globals__['run_sim'] = run_sim
  try:
    test_func()
  finally:
    test_func.__globals__['run_sim'] = _run_test

def test_normal_queue_stall():
  test_func = _normal_queue_stall
  _run_test = test_func.__globals__['run_sim']
  test_func.__globals__['run_sim'] = run_sim
  try:
    test_func()
  finally:
    test_func.__globals__['run_sim'] = _run_test

def test_pipe_queue():
  test_func = _pipe_queue
  _run_test = test_func.__globals__['run_sim']
  test_func.__globals__['run_sim'] = run_sim
  try:
    test_func()
  finally:
    test_func.__globals__['run_sim'] = _run_test

def test_pipe_queue_stall():
  test_func = _pipe_queue_stall
  _run_test = test_func.__globals__['run_sim']
  test_func.__globals__['run_sim'] = run_sim
  try:
    test_func()
  finally:
    test_func.__globals__['run_sim'] = _run_test

def test_bypass_queue():
  test_func = _bypass_queue
  _run_test = test_func.__globals__['run_sim']
  test_func.__globals__['run_sim'] = run_sim
  try:
    test_func()
  finally:
    test_func.__globals__['run_sim'] = _run_test

def test_bypass_queue_stall():
  test_func = _bypass_queue_stall
  _run_test = test_func.__globals__['run_sim']
  test_func.__globals__['run_sim'] = run_sim
  try:
    test_func()
  finally:
    test_func.__globals__['run_sim'] = _run_test



"""
==========================================================================
 xcel_ifcs.py
==========================================================================
Accelerator interface implementations at FL, CL, and RTL.

 Author: Yanghui Ou
   Date: June 3, 2019
"""
from greenlet import greenlet

from pymtl3 import *
from pymtl3.stdlib.connects import connect_pairs
from pymtl3.stdlib.rtl.queues import BypassQueueRTL, NormalQueueRTL

from .SendRecvIfc import RecvCL2SendRTL, RecvIfcRTL, RecvRTL2SendCL, SendIfcRTL
from .XcelMsg import XcelMsgType, mk_xcel_msg

#-------------------------------------------------------------------------
# FL interfaces
#-------------------------------------------------------------------------
# We assume FL interfaces shouldn't take types by default and the CL one should ...

class XcelMasterIfcFL( Interface ):

  def construct( s ):
    s.read  = CallerIfcFL()
    s.write = CallerIfcFL()

  def __str__( s ):
    return f"r{s.read}|w{s.write}"

  def connect( s, other, parent ):
    if isinstance( other, XcelMinionIfcRTL ):
      m = XcelIfcFL2RTLAdapter( other.ReqType, other.RespType )

      if hasattr( parent, "XcelIfcFL2RTL_count" ):
        count = parent.XcelIfcFL2RTL_count
        setattr( parent, "XcelIfcFL2RTL_" + str( count ), m )
      else:
        parent.XcelIfcFL2RTL_count = 0
        parent.XcelIfcFL2RTL_0 = m

      connect_pairs(
        s,       m.left,
        m.right, other,
      )
      parent.XcelIfcFL2RTL_count += 1
      return True

    elif isinstance( other, XcelMinionIfcCL ):
      m = XcelIfcFL2CLAdapter( other.ReqType, other.RespType )

      if hasattr( parent, "XcelIfcFL2CL_count" ):
        count = parent.XcelIfcFL2CL_count
        setattr( parent, "XcelIfcFL2CL_" + str( count ), m )
      else:
        parent.XcelIfcFL2CL_count = 0
        parent.XcelIfcFL2CL_0 = m

      connect_pairs(
        s,       m.left,
        m.right, other,
      )
      parent.XcelIfcFL2CL_count += 1
      return True

    return False

class XcelMinionIfcFL( Interface ):

  def construct( s, *, read=None, write=None ):
    s.read  = CalleeIfcFL( method=read )
    s.write = CalleeIfcFL( method=write )

  def __str__( s ):
    return f"r{s.read}|w{s.write}"

  def connect( s, other, parent ):
    if isinstance( other, XcelMasterIfcRTL ):
      m = XcelIfcRTL2FLAdapter( other.ReqType, other.RespType )

      if hasattr( parent, "XcelIfcRTL2FL_count" ):
        count = parent.XcelIfcRTL2FL_count
        setattr( parent, "XcelIfcRTL2FL_" + str( count ), m )
      else:
        parent.XcelIfcRTL2FL_count = 0
        parent.XcelIfcRTL2FL_0 = m

      connect_pairs(
        other,   m.left,
        m.right, s,
      )
      parent.XcelIfcRTL2FL_count += 1
      return True

    elif isinstance( other, XcelMasterIfcCL ):
      m = XcelIfcCL2FLAdapter( other.ReqType, other.RespType )

      if hasattr( parent, "XcelIfcCL2FL_count" ):
        count = parent.XcelIfcCL2FL_count
        setattr( parent, "XcelIfcCL2FL_" + str( count ), m )
      else:
        parent.XcelIfcCL2FL_count = 0
        parent.XcelIfcCL2FL_0 = m

      connect_pairs(
        other,   m.left,
        m.right, s,
      )
      parent.XcelIfcCL2FL_count += 1
      return True

    return False

  def line_trace( s ):
    return f"[r]{s.read}[w]{s.write}"

#-------------------------------------------------------------------------
# CL interfaces
#-------------------------------------------------------------------------

# There is no custom connect method in CL ifcs.
# For MasterCL-MinionRTL and MasterRTL-MinionCL connections, we just need
# to connect by name and leverage the custom connect method of the nested
# Send/RecvIfc. The CL-FL and FL-CL has been implemented in the FL ifc.

class XcelMasterIfcCL( Interface ):

  def construct( s, ReqType, RespType, resp=None, resp_rdy=None ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = CallerIfcCL( Type=ReqType )
    s.resp = CalleeIfcCL( Type=RespType, method=resp, rdy=resp_rdy )

  def __str__( s ):
    return "{},{}".format( s.req, s.resp )

class XcelMinionIfcCL( Interface ):

  def construct( s, ReqType, RespType, req=None, req_rdy=None ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = CalleeIfcCL( Type=ReqType, method=req, rdy=req_rdy )
    s.resp = CallerIfcCL( Type=RespType )

  def __str__( s ):
    return "{},{}".format( s.req, s.resp )

#-------------------------------------------------------------------------
# RTL interfaces
#-------------------------------------------------------------------------
# There is no custom connect method in CL ifcs.
# For MasterCL-MinionRTL and MasterRTL-MinionCL connections, we just need
# to connect by name and leverage the custom connect method of the nested
# Send/RecvIfc. The RTL-FL and FL-RTL has been implemented in the FL ifc.

class XcelMasterIfcRTL( Interface ):

  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = SendIfcRTL( ReqType  )
    s.resp = RecvIfcRTL( RespType )

  def __str__( s ):
    return "{},{}".format( s.req, s.resp )

class XcelMinionIfcRTL( Interface ):
  def construct( s, ReqType, RespType ):
    s.ReqType  = ReqType
    s.RespType = RespType
    s.req  = RecvIfcRTL( ReqType  )
    s.resp = SendIfcRTL( RespType )

  def __str__( s ):
    return "{},{}".format( s.req, s.resp )

#-------------------------------------------------------------------------
# CL/FL adapters
#-------------------------------------------------------------------------

class XcelIfcCL2FLAdapter( Component ):

  def recv_rdy( s ):
    return s.entry is None

  def recv( s, msg ):
    assert s.entry is None
    s.entry = msg

  def construct( s, ReqType, RespType ):
    s.left  = XcelMinionIfcCL( ReqType, RespType, req=s.recv, req_rdy=s.recv_rdy )
    s.right = XcelMasterIfcFL()
    s.entry = None

    @s.update
    def up_xcelifc_cl_fl_blk():

      if s.entry is not None and s.left.resp.rdy():

        # Dequeue xcel request message
        req     = s.entry
        s.entry = None

        if req.type_ == XcelMsgType.READ:
          resp = RespType( req.type_, s.right.read(req.addr) )

        elif req.type_ == XcelMsgType.WRITE:
          s.right.write( req.addr, req.data )
          resp = RespType( req.type_, 0 )

        # Make line trace look better since s.right might get blocked
        assert s.left.resp.rdy()
        s.left.resp( resp )

    s.add_constraints(
      M( s.left.req ) > U( up_xcelifc_cl_fl_blk ), # add an edge
    )

class XcelIfcFL2CLAdapter( Component ):

  def read( s, addr ):

    # TODO: refactor this greenlet stuff into some utility API
    while not s.right.req.rdy():
      greenlet.getcurrent().parent.switch(0)

    s.right.req( s.ReqType( XcelMsgType.READ, addr ) )

    while s.entry is None:
      greenlet.getcurrent().parent.switch(0)

    ret = s.entry.data
    s.entry = None
    return ret

  def write( s, addr, data ):

    while not s.right.req.rdy():
      greenlet.getcurrent().parent.switch(0)

    s.right.req( s.ReqType( XcelMsgType.WRITE, addr, data ) )

    while s.entry is None:
      greenlet.getcurrent().parent.switch(0)

    s.entry = None

  def recv_rdy( s ):
    return s.entry is None

  def recv( s, msg ):
    assert s.entry is None
    s.entry = msg

  def construct( s, ReqType, RespType ):
    s.entry = None # store response

    s.ReqType  = ReqType

    s.left  = XcelMinionIfcFL( read=s.read, write=s.write )
    s.right = XcelMasterIfcCL( ReqType, RespType, resp=s.recv, resp_rdy=s.recv_rdy )

    s.add_constraints(
      M( s.left.read  ) == M( s.right.req ),
      M( s.left.write ) == M( s.right.req ),
      M( s.left.read  ) > M( s.right.resp ),
      M( s.left.write ) > M( s.right.resp ),
    )

#-------------------------------------------------------------------------
# RTL/FL adapters
#-------------------------------------------------------------------------

class XcelIfcRTL2FLAdapter( Component ):

  def construct( s, ReqType, RespType ):
    s.left  = XcelMinionIfcRTL( ReqType, RespType )
    s.right = XcelMasterIfcFL()

    s.req_q = NormalQueueRTL( ReqType, num_entries=1 )
    connect( s.left.req, s.req_q.enq )

    @s.update
    def up_xcelifc_rtl_fl_blk():

      if s.req_q.deq.rdy and s.left.resp.rdy:

        if s.req_q.deq.ret.type_ == XcelMsgType.READ:
          resp = RespType( s.req_q.deq.ret.type_,
                           s.right.read( s.req_q.deq.ret.addr ) )

        elif s.req_q.deq.ret.type_ == XcelMsgType.WRITE:
          s.right.write( s.req_q.deq.ret.addr, s.req_q.deq.ret.data )
          resp = RespType( s.req_q.deq.ret.type_, 0 )

        s.req_q.deq.en = b1(1)
        s.left.resp.en  = b1(1)
        s.left.resp.msg = resp

      else:
        s.req_q.deq.en = b1(0)
        s.left.resp.en  = b1(0)

# Yanghui: directly adapting FL to RTL is tricky. I first convert FL to CL
# then CL to RTL using the adapters we already have.
class XcelIfcFL2RTLAdapter( Component ):

  def construct( s, ReqType, RespType ):
    s.left  = XcelMinionIfcFL ()
    s.right = XcelMasterIfcRTL( ReqType, RespType )

    s.fl2cl       = XcelIfcFL2CLAdapter( ReqType, RespType )
    s.req_cl2rtl  = RecvCL2SendRTL( ReqType )
    s.resp_rtl2cl = RecvRTL2SendCL( RespType)
    connect( s.left, s.fl2cl.left )
    connect_pairs(
      s.fl2cl.right.req, s.req_cl2rtl.recv,
      s.req_cl2rtl.send, s.right.req,
    )
    connect_pairs(
      s.fl2cl.right.resp, s.resp_rtl2cl.send,
      s.resp_rtl2cl.recv, s.right.resp,
    )



#=========================================================================
# StructuralRTLIRGenL1Pass.py
#=========================================================================
# Author : Shunning Jiang, Peitian Pan
# Date   : Apr 3, 2019
"""Provide L1 structural RTLIR generation pass."""

from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.rtype.RTLIRType import get_rtlir

from .StructuralRTLIRSignalExpr import gen_signal_expr


class StructuralRTLIRGenL1Pass( BasePass ):
  def __init__( s, inst_conns ):
    s.inst_conns = inst_conns

  def __call__( s, tr_top ):
    """ generate structural RTLIR for component `tr_top` """
    if not hasattr( tr_top, '_pass_structural_rtlir_gen' ):
      tr_top._pass_structural_rtlir_gen = PassMetadata()
    s.tr_top = tr_top
    try:
      s.gen_rtlir_types( tr_top )
      s.gen_constants( tr_top )
      s.sort_connections( tr_top )
    except AssertionError as e:
      msg = '' if not e.args is None else e.args[0]
      raise RTLIRConversionError( tr_top, msg )

  def gen_rtlir_types( s, tr_top ):
    tr_top._pass_structural_rtlir_gen.rtlir_type = get_rtlir( tr_top )

  def gen_constants( s, m ):
    ns = m._pass_structural_rtlir_gen
    ns.consts = []
    rtype = ns.rtlir_type
    const_types = rtype.get_consts_packed()
    for const_name, const_rtype in const_types:
      assert hasattr(m, const_name), \
        f"Internal error: {const_name} is not a member of {m}"
      const_instance = getattr(m, const_name)
      ns.consts.append( ( const_name, const_rtype, const_instance ) )

  def sort_connections( s, m ):
    m_conns_set   = s.inst_conns[m]
    ordered_conns = [ *m.get_connect_order() ]
    assert len(ordered_conns) == len(m_conns_set)

    for i, x in enumerate(ordered_conns):
      if x not in m_conns_set:
        x = (x[1], x[0])
        assert x in m_conns_set, "There is a connection missing from "\
                                 "connect_order. Please contact PyMTL developers!"
        ordered_conns[i] = x

    m._pass_structural_rtlir_gen.connections = \
      [ (gen_signal_expr(m, x[0]), gen_signal_expr(m, x[1])) for x in ordered_conns ]



#=========================================================================
# BehavioralRTLIRGenL5Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Oct 20, 2018
"""Provide L5 behavioral RTLIR generation pass."""

from pymtl3.passes.BasePass import BasePass, PassMetadata
from pymtl3.passes.rtlir.util.utility import get_ordered_upblks, get_ordered_update_ff

from .BehavioralRTLIRGenL4Pass import BehavioralRTLIRGeneratorL4


class BehavioralRTLIRGenL5Pass( BasePass ):
  def __call__( s, m ):
    """Generate RTLIR for all upblks of m."""
    if not hasattr( m, '_pass_behavioral_rtlir_gen' ):
      m._pass_behavioral_rtlir_gen = PassMetadata()
    m._pass_behavioral_rtlir_gen.rtlir_upblks = {}
    visitor = BehavioralRTLIRGeneratorL5( m )
    upblks = {
      'CombUpblk' : get_ordered_upblks(m),
      'SeqUpblk'  : get_ordered_update_ff(m),
    }
    # Sort the upblks by their name
    upblks['CombUpblk'].sort( key = lambda x: x.__name__ )
    upblks['SeqUpblk'].sort( key = lambda x: x.__name__ )

    for upblk_type in ( 'CombUpblk', 'SeqUpblk' ):
      for blk in upblks[ upblk_type ]:
        visitor._upblk_type = upblk_type
        upblk_info = m.get_update_block_info( blk )
        upblk = visitor.enter( blk, upblk_info[-1] )
        upblk.is_lambda = upblk_info[0]
        upblk.src       = upblk_info[1]
        upblk.lino      = upblk_info[2]
        upblk.filename  = upblk_info[3]
        m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ] = upblk

class BehavioralRTLIRGeneratorL5( BehavioralRTLIRGeneratorL4 ):
  """Behavioral RTLIR generator level 5.

  Do nothing here because attributes have been handled in previous
  levels.
  """
  def __init__( s, component ):
    super().__init__( component )



#=========================================================================
# SVNameMangle_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 30, 2019
"""Test the SystemVerilog name mangling."""

from pymtl3.datatypes import Bits1, Bits32, bitstruct
from pymtl3.dsl import Component, InPort, Interface, OutPort
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.util.test_utility import do_test

from ..ImportPass import ImportPass


def local_do_test( m ):
  m.elaborate()
  rtype = rt.get_component_ifc_rtlir( m )
  ipass = ImportPass()
  result = ipass.gen_packed_ports( rtype )
  assert result == m._ref_ports

def test_port_single( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
  a = A()
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in_', rt.Port('input', rdt.Vector(32)) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = a._ref_ports
  do_test( a )

def test_port_array( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range( 3 ) ]
  a = A()
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in_', rt.Array([3], rt.Port('input', rdt.Vector(32))) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in___0', rt.Port('input', rdt.Vector(32)) ),
    ( 'in___1', rt.Port('input', rdt.Vector(32)) ),
    ( 'in___2', rt.Port('input', rdt.Vector(32)) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  do_test( a )

def test_port_2d_array( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ [ InPort( Bits32 ) for _ in range(2) ] for _ in range(3) ]
  a = A()
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in_', rt.Array( [3, 2], rt.Port('input', rdt.Vector(32)) ) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in___0__0', rt.Port('input', rdt.Vector(32)) ),
    ( 'in___0__1', rt.Port('input', rdt.Vector(32)) ),
    ( 'in___1__0', rt.Port('input', rdt.Vector(32)) ),
    ( 'in___1__1', rt.Port('input', rdt.Vector(32)) ),
    ( 'in___2__0', rt.Port('input', rdt.Vector(32)) ),
    ( 'in___2__1', rt.Port('input', rdt.Vector(32)) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  do_test( a )

def test_struct_port_single( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( struct )
  a = A()
  st = rdt.Struct('struct', {'bar':rdt.Vector(32), 'foo':rdt.Vector(32)})
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in_', rt.Port('input', st ) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in___bar', rt.Port('input', rdt.Vector(32) ) ),
    ( 'in___foo', rt.Port('input', rdt.Vector(32) ) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  do_test( a )

def test_struct_port_array( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  st = rdt.Struct('struct', {'bar':rdt.Vector(32), 'foo':rdt.Vector(32)})
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in_', rt.Array([2], rt.Port('input', st)) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in___0__bar', rt.Port('input', rdt.Vector(32) ) ),
    ( 'in___0__foo', rt.Port('input', rdt.Vector(32) ) ),
    ( 'in___1__bar', rt.Port('input', rdt.Vector(32) ) ),
    ( 'in___1__foo', rt.Port('input', rdt.Vector(32) ) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  do_test( a )

def test_packed_array_port_array( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: [ [ Bits32 ] * 2 ] * 3
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  foo = rdt.PackedArray([3,2], rdt.Vector(32))
  st = rdt.Struct('struct', {'bar':rdt.Vector(32), 'foo':foo})
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in_', rt.Array([2], rt.Port('input', st ))),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in___0__bar', rt.Port('input', rdt.Vector(32) )),
    ( 'in___0__foo__0__0', rt.Port('input', rdt.Vector(32) )),
    ( 'in___0__foo__0__1', rt.Port('input', rdt.Vector(32) )),
    ( 'in___0__foo__1__0', rt.Port('input', rdt.Vector(32) )),
    ( 'in___0__foo__1__1', rt.Port('input', rdt.Vector(32) )),
    ( 'in___0__foo__2__0', rt.Port('input', rdt.Vector(32) )),
    ( 'in___0__foo__2__1', rt.Port('input', rdt.Vector(32) )),
    ( 'in___1__bar', rt.Port('input', rdt.Vector(32) )),
    ( 'in___1__foo__0__0', rt.Port('input', rdt.Vector(32) )),
    ( 'in___1__foo__0__1', rt.Port('input', rdt.Vector(32) )),
    ( 'in___1__foo__1__0', rt.Port('input', rdt.Vector(32) )),
    ( 'in___1__foo__1__1', rt.Port('input', rdt.Vector(32) )),
    ( 'in___1__foo__2__0', rt.Port('input', rdt.Vector(32) )),
    ( 'in___1__foo__2__1', rt.Port('input', rdt.Vector(32) )),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  do_test( a )

def test_nested_struct( do_test ):
  @bitstruct
  class inner_struct:
    foo: Bits32
  @bitstruct
  class struct:
    bar: Bits32
    inner: inner_struct
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  inner = rdt.Struct('inner_struct', {'foo':rdt.Vector(32)})
  st = rdt.Struct('struct', {'bar':rdt.Vector(32), 'inner':inner})
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in_', rt.Array([2], rt.Port('input', st )) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'in___0__bar', rt.Port('input', rdt.Vector(32) ) ),
    ( 'in___0__inner__foo', rt.Port('input', rdt.Vector(32) ) ),
    ( 'in___1__bar', rt.Port('input', rdt.Vector(32) ) ),
    ( 'in___1__inner__foo', rt.Port('input', rdt.Vector(32) ) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) )
  ]
  do_test( a )

def test_interface( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = Ifc()
  a = A()
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) ),
    ( 'ifc__msg', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__rdy', rt.Port('output', rdt.Vector(1)) ),
    ( 'ifc__val', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = a._ref_ports
  do_test( a )

def test_interface_array( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) ),
    ( 'ifc__0__msg', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__0__rdy', rt.Port('output', rdt.Vector(1)) ),
    ( 'ifc__0__val', rt.Port('input', rdt.Vector(1)) ),
    ( 'ifc__1__msg', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__1__rdy', rt.Port('output', rdt.Vector(1)) ),
    ( 'ifc__1__val', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = a._ref_ports
  do_test( a )

def test_nested_interface( do_test ):
  class InnerIfc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class Ifc( Interface ):
    def construct( s ):
      s.valrdy_ifc = InnerIfc()
      s.ctrl_bar = InPort( Bits32 )
      s.ctrl_foo = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) ),
    ( 'ifc__0__ctrl_bar', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__0__ctrl_foo', rt.Port('output', rdt.Vector(32)) ),
    ( 'ifc__0__valrdy_ifc__msg', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__0__valrdy_ifc__rdy', rt.Port('output', rdt.Vector(1)) ),
    ( 'ifc__0__valrdy_ifc__val', rt.Port('input', rdt.Vector(1)) ),
    ( 'ifc__1__ctrl_bar', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__1__ctrl_foo', rt.Port('output', rdt.Vector(32)) ),
    ( 'ifc__1__valrdy_ifc__msg', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__1__valrdy_ifc__rdy', rt.Port('output', rdt.Vector(1)) ),
    ( 'ifc__1__valrdy_ifc__val', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = a._ref_ports
  do_test( a )

def test_nested_interface_port_array( do_test ):
  class InnerIfc( Interface ):
    def construct( s ):
      s.msg = [ InPort( Bits32 ) for _ in range(2) ]
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class Ifc( Interface ):
    def construct( s ):
      s.valrdy_ifc = InnerIfc()
      s.ctrl_bar = InPort( Bits32 )
      s.ctrl_foo = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  a._ref_ports = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) ),
    ( 'ifc__0__ctrl_bar', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__0__ctrl_foo', rt.Port('output', rdt.Vector(32)) ),
    ( 'ifc__0__valrdy_ifc__msg', rt.Array([2], rt.Port('input', rdt.Vector(32))) ),
    ( 'ifc__0__valrdy_ifc__rdy', rt.Port('output', rdt.Vector(1)) ),
    ( 'ifc__0__valrdy_ifc__val', rt.Port('input', rdt.Vector(1)) ),
    ( 'ifc__1__ctrl_bar', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__1__ctrl_foo', rt.Port('output', rdt.Vector(32)) ),
    ( 'ifc__1__valrdy_ifc__msg', rt.Array([2], rt.Port('input', rdt.Vector(32))) ),
    ( 'ifc__1__valrdy_ifc__rdy', rt.Port('output', rdt.Vector(1)) ),
    ( 'ifc__1__valrdy_ifc__val', rt.Port('input', rdt.Vector(1)) )
  ]
  a._ref_ports_yosys = [
    ( 'clk', rt.Port('input', rdt.Vector(1)) ),
    ( 'reset', rt.Port('input', rdt.Vector(1)) ),
    ( 'ifc__0__ctrl_bar', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__0__ctrl_foo', rt.Port('output', rdt.Vector(32)) ),
    ( 'ifc__0__valrdy_ifc__msg__0', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__0__valrdy_ifc__msg__1', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__0__valrdy_ifc__rdy', rt.Port('output', rdt.Vector(1)) ),
    ( 'ifc__0__valrdy_ifc__val', rt.Port('input', rdt.Vector(1)) ),
    ( 'ifc__1__ctrl_bar', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__1__ctrl_foo', rt.Port('output', rdt.Vector(32)) ),
    ( 'ifc__1__valrdy_ifc__msg__0', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__1__valrdy_ifc__msg__1', rt.Port('input', rdt.Vector(32)) ),
    ( 'ifc__1__valrdy_ifc__rdy', rt.Port('output', rdt.Vector(1)) ),
    ( 'ifc__1__valrdy_ifc__val', rt.Port('input', rdt.Vector(1)) )
  ]
  do_test( a )



"""
========================================================================
enrdy_queues_test.py
========================================================================

Author : Shunning Jiang
Date   : Mar 9, 2018
"""
import pytest

from pymtl3 import *
from pymtl3.stdlib.test import TestSinkCL, TestSrcCL

from .enrdy_queues import *

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, Type, q, src_msgs, sink_msgs, src_interval, sink_interval ):

    # Instantiate models

    s.src  = TestSrcCL( Type, src_msgs, interval_delay=src_interval )
    s.q    = q
    s.sink = TestSinkCL( Type, sink_msgs, interval_delay=sink_interval )

    # Connect

    connect( s.src.send, s.q.enq )
    connect( s.sink.recv, s.q.deq )

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace(s ):
    return s.src.line_trace() + " " + s.q.line_trace() + " " + s.sink.line_trace()

def run_sim( model ):
  model.apply( SimulationPass() )

  print()
  cycle = 0
  while not model.done() and cycle < 1000:
    model.tick()
    print( "{:3}: {}".format( cycle, model.line_trace() ) )
    cycle += 1

  assert cycle < 1000

  model.tick()
  model.tick()
  model.tick()


F = lambda x: Bits32(x)

req  = list(map( F, [1,2,3,4,5,6,7,8,9,10] ))
resp = list(map( F, [1,2,3,4,5,6,7,8,9,10] ))

def test_normal_queue():
  run_sim( TestHarness( Bits32, NormalQueue1RTL(Bits32), req, resp, 0, 0 ) )

def test_normal_queue_stall():
  run_sim( TestHarness( Bits32, NormalQueue1RTL(Bits32), req, resp, 3, 4 ) )

def test_pipe_queue():
  run_sim( TestHarness( Bits32, PipeQueue1RTL(Bits32), req, resp, 0, 0 ) )

def test_pipe_queue_stall():
  run_sim( TestHarness( Bits32, PipeQueue1RTL(Bits32), req, resp, 3, 4  ) )

def test_bypass_queue():
  run_sim( TestHarness( Bits32, BypassQueue1RTL(Bits32), req, resp, 0, 0 ) )

def test_bypass_queue_stall():
  run_sim( TestHarness( Bits32, BypassQueue1RTL(Bits32), req, resp, 3, 4  ) )

def test_bypass_queue2():
  run_sim( TestHarness( Bits32, BypassQueue2RTL(Bits32), req, resp, 0, 0 ) )

def test_bypass_queue2_stall():
  run_sim( TestHarness( Bits32, BypassQueue2RTL(Bits32), req, resp, 3, 4  ) )



#=========================================================================
# BehavioralTranslatorL1.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 22, 2019
"""Provide L1 behavioral translator."""


from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL1Pass import (
    BehavioralRTLIRGenL1Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL1Pass import (
    BehavioralRTLIRTypeCheckL1Pass,
)
from pymtl3.passes.rtlir.util.utility import get_ordered_upblks, get_ordered_update_ff

from .BehavioralTranslatorL0 import BehavioralTranslatorL0


class BehavioralTranslatorL1( BehavioralTranslatorL0 ):
  def __init__( s, top ):
    super().__init__( top )

  def clear( s, tr_top ):
    super().clear( tr_top )
    s.gen_behavioral_trans_metadata( tr_top )

  #-----------------------------------------------------------------------
  # gen_behavioral_trans_metadata
  #-----------------------------------------------------------------------

  def gen_behavioral_trans_metadata( s, tr_top ):
    s.behavioral.rtlir = {}
    s.behavioral.freevars = {}
    s.behavioral.accessed = {}
    s.behavioral.upblk_decls = {}
    s.behavioral.upblk_srcs = {}
    s.behavioral.upblk_py_srcs = {}
    s.behavioral.decl_freevars = {}
    s._gen_behavioral_trans_metadata( tr_top )

  #-----------------------------------------------------------------------
  # _gen_behavioral_trans_metadata
  #-----------------------------------------------------------------------

  def _gen_behavioral_trans_metadata( s, m ):
    m.apply( BehavioralRTLIRGenL1Pass() )
    m.apply( BehavioralRTLIRTypeCheckL1Pass() )
    s.behavioral.rtlir[m] = \
        m._pass_behavioral_rtlir_gen.rtlir_upblks
    s.behavioral.freevars[m] = \
        m._pass_behavioral_rtlir_type_check.rtlir_freevars

  #-----------------------------------------------------------------------
  # translate_behavioral
  #-----------------------------------------------------------------------

  # Override
  def translate_behavioral( s, m ):
    """Translate behavioral part of `m`."""
    # Get upblk metadata
    s.behavioral.accessed[m] = m._pass_behavioral_rtlir_type_check.rtlir_accessed
    # Translate upblks
    upblk_decls = []
    upblk_srcs = []
    upblk_py_srcs = []
    upblks = {
      'CombUpblk' : get_ordered_upblks(m),
      'SeqUpblk'  : get_ordered_update_ff(m),
    }
    # Sort the upblks by their name
    upblks['CombUpblk'].sort( key = lambda x: x.__name__ )
    upblks['SeqUpblk'].sort( key = lambda x: x.__name__ )

    for upblk_type in ( 'CombUpblk', 'SeqUpblk' ):
      for blk in upblks[ upblk_type ]:
        upblk_ir = s.behavioral.rtlir[ m ][ blk ]
        upblk_srcs.append( s.rtlir_tr_upblk_src(
          blk, upblk_ir
        ) )
        upblk_py_srcs.append( s.rtlir_tr_upblk_py_src(
          blk,
          upblk_ir.is_lambda,
          upblk_ir.src,
          upblk_ir.lino,
          upblk_ir.filename
        ) )
        upblk_decls.append( s.rtlir_tr_upblk_decl(
          blk, upblk_srcs[-1], upblk_py_srcs[-1]
        ) )
    s.behavioral.upblk_srcs[m] = s.rtlir_tr_upblk_srcs( upblk_srcs )
    s.behavioral.upblk_py_srcs[m] = s.rtlir_tr_upblk_decls( upblk_py_srcs )
    s.behavioral.upblk_decls[m] = s.rtlir_tr_upblk_decls( upblk_decls )

    # Generate free variable declarations
    freevars = []
    for name, fvar in s.behavioral.freevars[m].items():
      rtype = rt.get_rtlir( fvar )
      if isinstance( rtype, rt.Array ):
        fvar_rtype = rtype.get_sub_type()
        array_rtype = rtype
      else:
        fvar_rtype = rtype
        array_rtype = None
      dtype = fvar_rtype.get_dtype()
      assert isinstance( dtype, rdt.Vector ), \
        f'{name} freevar should be an integer or a list of integers!'
      freevars.append( s.rtlir_tr_behavioral_freevar(
        name,
        fvar_rtype,
        s.rtlir_tr_unpacked_array_type( array_rtype ),
        s.rtlir_tr_vector_dtype( dtype ),
        fvar
      ) )
    s.behavioral.decl_freevars[m] = s.rtlir_tr_behavioral_freevars(freevars)

  #-----------------------------------------------------------------------
  # Methods to be implemented by the backend translator
  #-----------------------------------------------------------------------

  def rtlir_tr_upblk_decls( s, upblk_decls ):
    raise NotImplementedError()

  def rtlir_tr_upblk_decl( s, upblk, src, py_src ):
    raise NotImplementedError()

  def rtlir_tr_upblk_srcs( s, upblk_srcs ):
    raise NotImplementedError()

  def rtlir_tr_upblk_src( s, upblk, rtlir_upblk ):
    raise NotImplementedError()

  def rtlir_tr_upblk_py_srcs( s, upblk_py_srcs ):
    raise NotImplementedError()

  def rtlir_tr_upblk_py_src( s, upblk, is_lambda, src, lino, filename ):
    raise NotImplementedError()

  def rtlir_tr_behavioral_freevars( s, freevars ):
    raise NotImplementedError()

  def rtlir_tr_behavioral_freevar( s, id_, rtype, array_type, dtype, obj ):
    raise NotImplementedError()
